#!/usr/bin/env python3
"""Generate a Python client from Swagger specification."""

# pylint: disable=missing-docstring,too-many-instance-attributes,too-many-locals,too-many-ancestors,too-many-branches
# pylint: disable=too-many-statements,too-many-lines

import collections
from typing import MutableMapping, Union, List, Optional, Dict  # pylint: disable=unused-import

import icontract
import jinja2
import jinja2.exceptions

import swagger_to
import swagger_to.intermediate
import swagger_to.swagger


class Typedef:
    """Represent a type definition in Python."""

    def __init__(self) -> None:
        """Initialize with defaults."""
        self.description = ''
        self.identifier = ''

    def __str__(self) -> str:
        """Represent the type definition as its class name and the identifier."""
        return '{}({})'.format(self.__class__.__name__, self.identifier)


class Booldef(Typedef):
    """Represent a Python boolean."""

    pass


class Intdef(Typedef):
    """Represent a Python integer."""

    pass


class Floatdef(Typedef):
    """Represent a Python floating-point number."""

    pass


class Strdef(Typedef):
    """Represent a Python string."""

    pass


class Bytesdef(Typedef):
    """Represent a Python immutable bytes object."""

    pass


class Filedef(Typedef):
    """Represents a Python file type in form-data parameters."""

    pass


class Attribute:
    """Represent an instance attribute of a Python class."""

    def __init__(self) -> None:
        """Initialize with defaults."""
        self.typedef = None  # type: Optional[Typedef]
        self.description = ''
        self.name = ''
        self.required = False

        self.classdef = None  # type: Optional[Classdef]


class Classdef(Typedef):
    """Represent a definition of a Python class."""

    def __init__(self) -> None:
        """Initialize with defaults."""
        super().__init__()
        self.attributes = collections.OrderedDict()  # type: MutableMapping[str, Attribute]


class Listdef(Typedef):
    """Represent a definition of a Python list."""

    def __init__(self) -> None:
        """Initialize with defaults."""
        super().__init__()

        self.items = None  # type: Union[None, Typedef]


class Dictdef(Typedef):
    """Represent a definition of a Python dictionary."""

    def __init__(self) -> None:
        """Initialize with defaults."""
        super().__init__()

        self.values = None  # type: Union[None, Typedef]


class Parameter:
    """Represent a parameter of a request in Python."""

    def __init__(self) -> None:
        """Initialize with defaults."""
        # Original parameter name
        self.name = ''

        # Parameter identifier in the code
        self.identifier = ''

        self.in_what = ''
        self.typedef = None  # type: Optional[Typedef]
        self.required = False
        self.description = None  # type: Optional[str]


class Response:
    """Represent a response from the server in Python."""

    def __init__(self) -> None:
        """Initialize with defaults."""
        self.code = ''
        self.typedef = None  # type: Optional[Typedef]
        self.description = ''


class Request:
    """Represent a request function of the client in Python."""

    def __init__(self) -> None:
        """Initialize with defaults."""
        self.operation_id = ''
        self.path = ''
        self.method = ''
        self.description = ''

        # the union of all parameters
        self.parameters = []  # type: List[Parameter]

        # parameters split into categories
        self.body_parameter = None  # type: Optional[Parameter]
        self.header_parameters = []  # type: List[Parameter]
        self.query_parameters = []  # type: List[Parameter]
        self.path_parameters = []  # type: List[Parameter]
        self.formdata_parameters = []  # type: List[Parameter]
        self.file_parameters = []  # type: List[Parameter]

        self.responses = collections.OrderedDict()  # type: MutableMapping[str, Response]
        self.produces = []  # type: List[str]


def _anonymous_or_get_typedef(intermediate_typedef: swagger_to.intermediate.Typedef,
                              typedefs: MutableMapping[str, Typedef]) -> Typedef:
    """
    Get the Python representation of the type definition from the table of Python type definitions by its identifier.

    If the type definition has no identifier, it is translated on the spot to the corresponding Python type definition.

    :param intermediate_typedef: intermediate representation of the type definition
    :param typedefs: table of type definitions in Python representation
    :return: type definition in Python representation
    """
    if intermediate_typedef.identifier:
        if intermediate_typedef.identifier not in typedefs:
            raise KeyError('Intermediate typedef not found in the translated typedefs: {!r}'.format(
                intermediate_typedef.identifier))

        return typedefs[intermediate_typedef.identifier]

    return _to_typedef(intermediate_typedef=intermediate_typedef)


def _to_typedef(intermediate_typedef: swagger_to.intermediate.Typedef) -> Typedef:
    """
    Translate the type definition in intermediate representation to Python.

    :param intermediate_typedef: type definition in intermediate representation
    :return: type definition in Python representation
    """
    typedef = None  # type: Optional[Typedef]

    if isinstance(intermediate_typedef, swagger_to.intermediate.Primitivedef):
        if intermediate_typedef.type == 'boolean':
            typedef = Booldef()
        elif intermediate_typedef.type == 'integer':
            typedef = Intdef()
        elif intermediate_typedef.type == 'number':
            typedef = Floatdef()
        elif intermediate_typedef.type == 'string':
            typedef = Strdef()
        elif intermediate_typedef.type == 'file':
            typedef = Filedef()
        else:
            raise NotImplementedError('Converting intermediate type to Python is not supported: {}'.format(
                intermediate_typedef.type))

    elif isinstance(intermediate_typedef, swagger_to.intermediate.Arraydef):
        typedef = Listdef()
        typedef.items = _to_typedef(intermediate_typedef=intermediate_typedef.items)

    elif isinstance(intermediate_typedef, swagger_to.intermediate.Mapdef):
        typedef = Dictdef()
        typedef.values = _to_typedef(intermediate_typedef=intermediate_typedef.values)

    elif isinstance(intermediate_typedef, swagger_to.intermediate.Objectdef):
        typedef = Classdef()

        for intermediate_prop in intermediate_typedef.properties.values():
            attr = Attribute()
            attr.description = intermediate_prop.description
            attr.name = intermediate_prop.name
            attr.typedef = _to_typedef(intermediate_typedef=intermediate_prop.typedef)
            attr.required = intermediate_prop.required
            attr.classdef = typedef

            typedef.attributes[attr.name] = attr

        typedef.attributes = collections.OrderedDict(
            sorted(list(typedef.attributes.items()), key=lambda kv: not kv[1].required))
    else:
        raise NotImplementedError('Converting intermediate typedef to Python is not supported: {!r}'.format(
            type(intermediate_typedef)))

    typedef.description = intermediate_typedef.description
    typedef.identifier = intermediate_typedef.identifier

    assert typedef is not None

    return typedef


def to_typedefs(
        intermediate_typedefs: MutableMapping[str, swagger_to.intermediate.Typedef]) -> MutableMapping[str, Typedef]:
    """
    Translate type definitions from intermediate representation to Python.

    :param intermediate_typedefs: table of type definitions in intermediate representation
    :return: table of type definitions in Python representation
    """
    typedefs = collections.OrderedDict()  # type: MutableMapping[str, Typedef]

    for intermediate_typedef in intermediate_typedefs.values():
        assert intermediate_typedef is not None

        typedef = _to_typedef(intermediate_typedef=intermediate_typedef)
        typedefs[typedef.identifier] = typedef

    return typedefs


def _to_response(intermediate_response: swagger_to.intermediate.Response,
                 typedefs: MutableMapping[str, Typedef]) -> Response:
    """
    Translate an endpoint response from the intermediate to a Python representation.

    :param intermediate_response: intermediate representation of a response
    :param typedefs: table of type definitions in Python representation
    :return: Python representation of the response
    """
    resp = Response()
    resp.code = intermediate_response.code
    resp.typedef = None if intermediate_response.typedef is None else \
        _anonymous_or_get_typedef(intermediate_typedef=intermediate_response.typedef, typedefs=typedefs)
    resp.description = intermediate_response.description
    return resp

@icontract.ensure(
    lambda result:
    sorted(result.parameters, key=id) == sorted((
        param
        for param in ([result.body_parameter] if result.body_parameter else []) +
                     result.query_parameters +
                     result.header_parameters +
                     result.path_parameters +
                     result.formdata_parameters +
                     result.file_parameters), key=id),
    enabled=icontract.SLOW)
@icontract.ensure(lambda result: all(isinstance(param.typedef, Filedef) for param in result.file_parameters))
def _to_request(endpoint: swagger_to.intermediate.Endpoint, typedefs: MutableMapping[str, Typedef]) -> Request:
    """
    Translate an endpoint from an intermediate representation to a Python client request function.

    :param endpoint: intermediate representation of the endpoint
    :param typedefs: table of type definitions in Python representation
    :return: Python representation of the client request function
    """
    req = Request()
    req.description = endpoint.description
    req.method = endpoint.method
    req.operation_id = endpoint.operation_id
    req.path = endpoint.path

    ##
    # Generate identifiers corresponding to the parameters.
    ##

    for intermediate_param in endpoint.parameters:
        if intermediate_param.name == "":
            raise ValueError("Unexpected empty intermediate parameter name in the endpoint: {}".format(
                endpoint.operation_id))

    param_to_identifier = {
        intermediate_param: swagger_to.snake_case(identifier=intermediate_param.name)
        for intermediate_param in endpoint.parameters
    }

    identifiers = list(param_to_identifier.values())
    needs_location_prefix = len(set(identifiers)) != len(identifiers)
    if needs_location_prefix:
        param_to_identifier = {
            inter_param: swagger_to.snake_case(identifier="{}_{}".format(inter_param.in_what, inter_param.name))
            for inter_param in endpoint.parameters
        }

        ##
        # Assert that there are no conflicts at this point
        ##

        by_identifier = collections.defaultdict(
            list)  # type: MutableMapping[str, List[swagger_to.intermediate.Parameter]]
        for inter_param, identifier in param_to_identifier.items():
            by_identifier[identifier].append(inter_param)

        # yapf: disable
        msgs = [
            "in the endpoint {} {} for the identifier {!r}: {}".format(
                endpoint.method.upper(), endpoint.path, identifier, ", ".join(
                    ["{} in {}".format(inter_param.name, inter_param.in_what)
                     for inter_param in inter_params]))
            for identifier, inter_params in by_identifier.items()
            if len(inter_params) > 1
        ]
        # yapf: enable

        if len(msgs) > 0:
            raise ValueError("There are conflicting identifiers for parameters:\n{}".format("\n".join(msgs)))

    ##
    # Convert parameters
    ##

    assert all(intermediate_param in param_to_identifier for intermediate_param in endpoint.parameters), \
        "Expected all parameters to have a generated argument identifier."

    req.parameters = []
    for intermediate_param in endpoint.parameters:
        identifier = param_to_identifier[intermediate_param]

        param = Parameter()
        param.identifier = identifier
        param.name = intermediate_param.name
        param.in_what = intermediate_param.in_what
        param.typedef = _anonymous_or_get_typedef(intermediate_typedef=intermediate_param.typedef, typedefs=typedefs)
        param.required = intermediate_param.required
        param.description = intermediate_param.description

        req.parameters.append(param)

    ##
    # Check that there is only one body parameter
    ##

    body_parameter_names = [param.name for param in req.parameters if param.in_what == 'body']
    if len(body_parameter_names) > 1:
        raise KeyError('Duplicate body parameters in an endpoint: {}'.format(body_parameter_names))

    ##
    # Categorize parameters
    ##

    for param in req.parameters:
        if isinstance(param.typedef, Filedef):
            req.file_parameters.append(param)
        elif param.in_what == 'body':
            req.body_parameter = param
        elif param.in_what == 'header':
            req.header_parameters.append(param)
        elif param.in_what == 'query':
            req.query_parameters.append(param)
        elif param.in_what == 'path':
            req.path_parameters.append(param)
        elif param.in_what == 'formData':
            req.formdata_parameters.append(param)
        else:
            raise NotImplementedError("Unsupported parameter location: in {}".format(param.in_what))

    # Parameters are sorted so that first come the required ones; python requires the optional parameters to come
    # at the end.
    req.parameters.sort(key=lambda param: not param.required)

    ##
    # Convert responses
    ##

    for code, intermediate_resp in endpoint.responses.items():
        req.responses[code] = _to_response(intermediate_response=intermediate_resp, typedefs=typedefs)

    req.produces = endpoint.produces[:]

    return req


@icontract.ensure(lambda endpoints, result: len(endpoints) == len(result))
def to_requests(endpoints: List[swagger_to.intermediate.Endpoint],
                typedefs: MutableMapping[str, Typedef]) -> List[Request]:
    """
    Translate endpoints from intermediate representation to Python client request functions.

    :param endpoints: intermediate representation of the endpoints
    :param typedefs: table of type definitions in Python representation
    :return: Python representation of client's request functions corresponding to the endpoints
    """
    requests = []  # type: List[Request]
    for endpoint in endpoints:
        requests.append(_to_request(endpoint=endpoint, typedefs=typedefs))

    return requests


@icontract.ensure(lambda result: not result.endswith('\n'))
def _comment(text: str) -> str:
    r"""
    Generate a (possibly multi-line) comment from the text.

    >>> cmt = _comment('  testme\n  \nagain\n')
    >>> assert cmt == '#   testme\n#\n# again\n#'

    :param text: of the comment
    :return: Python code
    """
    out = []  # type: List[str]
    lines = text.split('\n')
    for line in lines:
        rstripped = line.rstrip()
        if len(rstripped) > 0:
            out.append('# {}'.format(rstripped))
        else:
            out.append('#')

    return '\n'.join(out)


@icontract.require(lambda text: text != '')
@icontract.ensure(lambda result: not result.endswith('\n'))
@icontract.ensure(lambda result: result.startswith('r"""') or result.startswith('"""'))
@icontract.ensure(lambda result: result.endswith('"""'))
def _docstring(text: str) -> str:
    """
    Generate a (possibly multi-line) docstring from the text.

    :param text: of the docstring
    :return: Python code
    """
    has_backslash = '\\' in text
    has_triple_quote = '"""' in text

    is_raw = False
    content = text

    if has_triple_quote:
        is_raw = False
        content = text.replace('\\', '\\\\').replace('"""', '\\"\\"\\"')
    else:
        if has_backslash:
            is_raw = True

    parts = []  # type: List[str]
    if is_raw:
        parts.append('r')

    lines = text.splitlines()
    if len(lines) > 1:
        parts.append('"""\n')
        parts.append(content)
        parts.append('\n"""')
    elif len(lines) == 1:
        parts.append('"""{}"""'.format(content))

    return ''.join(parts)


def _raise(message: str) -> None:
    """Raise an exception in a template."""
    raise Exception(message)


# Jinja2 environment
_ENV = jinja2.Environment(trim_blocks=True, lstrip_blocks=True, loader=jinja2.BaseLoader())
_ENV.filters.update({
    'comment': _comment,
    'docstring': _docstring,
    'snake_case': swagger_to.snake_case,
    'raise': _raise,
    'repr': lambda value: repr(value),
    'upper_first': swagger_to.upper_first
})


def _from_string_with_informative_exceptions(env: jinja2.Environment, text: str) -> jinja2.Template:
    """
    Parse the jinja2 template raising more informative exceptions if there are any.

    :param env: global jinja2 environment
    :param text: text of the template
    :return: parsed template
    """
    syntax_err = None  # type: Optional[jinja2.exceptions.TemplateSyntaxError]
    try:
        return env.from_string(source=text)
    except jinja2.exceptions.TemplateSyntaxError as err:
        syntax_err = err

    if syntax_err is not None:
        lines = text.splitlines()
        line = lines[syntax_err.lineno - 1]

        msg = '{}\n{}'.format(syntax_err.message, line)

        raise jinja2.exceptions.TemplateSyntaxError(
            message=msg, lineno=syntax_err.lineno, name=syntax_err.name, filename=syntax_err.filename)
    else:
        raise AssertionError("Unhandled execution path")


def _type_expression(typedef: Typedef, path: Optional[str] = None) -> str:
    """
    Translate the type definition in Python representation to a type expression as Python code.

    :param typedef: Python representation of the type definition
    :param path: path in the Swagger spec
    :return: Python code
    """
    # pylint: disable=too-many-return-statements
    if isinstance(typedef, Booldef):
        return 'bool'
    elif isinstance(typedef, Intdef):
        return 'int'
    elif isinstance(typedef, Floatdef):
        return 'float'
    elif isinstance(typedef, Strdef):
        return 'str'
    elif isinstance(typedef, Bytesdef):
        return 'bytes'
    elif isinstance(typedef, Filedef):
        return 'BinaryIO'
    elif isinstance(typedef, Listdef):
        if typedef.items is None:
            raise ValueError('Unexpected None items in typedef: {!r}'.format(typedef.identifier))

        return 'List[' + _type_expression(typedef=typedef.items, path=str(path) + '.items') + ']'
    elif isinstance(typedef, Dictdef):
        if typedef.values is None:
            raise ValueError('Unexpected None values in typedef: {!r}'.format(typedef.identifier))

        return 'Dict[str, ' + _type_expression(typedef=typedef.values, path=str(path) + '.values') + ']'
    elif isinstance(typedef, Classdef):
        if typedef.identifier == '':
            raise NotImplementedError(('Translating an anonymous class to a type expression '
                                       'is not supported: {}').format(path))

        return typedef.identifier
    else:
        raise NotImplementedError('Translating the typedef to a type expression is not supported: {!r}: {}'.format(
            type(typedef), path))


_CLASS_DEF_WO_ATTRIBUTES = _from_string_with_informative_exceptions(
    env=_ENV,
    text='''\
{% if classdef.attributes %}{{ raise('Expected a classdef without attributes, but got some.') }}{% endif %}
class {{ classdef.identifier }}:
    {% if classdef.description %}
    {{ classdef.description|upper_first|docstring|indent }}

    {% endif %}
    def to_jsonable(self) -> MutableMapping[str, Any]:
        """
        Dispatches the conversion to {{ classdef.identifier|snake_case }}_to_jsonable.

        :return: a JSON-able representation
        """
        return {{ classdef.identifier|snake_case }}_to_jsonable(self)''')

_CLASS_DEF_WITH_ATTRIBUTES_TPL = _from_string_with_informative_exceptions(
    env=_ENV,
    text='''\
{% if not classdef.attributes %}{{ raise('Expected a classdefinition with attributes, but got none.') }}{% endif %}
class {{ classdef.identifier }}:
    {% if classdef.description %}
    {{ classdef.description|upper_first|docstring|indent }}

    {% endif %}
    def __init__(
            self,
    {% for attr in classdef.attributes.values() %}
    {% if not attr.required %}
            {{ attr.name }}: Optional[{{ attribute_type[attr] }}] = None{{ ') -> None:' if loop.last else ',' }}
    {% else %}
            {{ attr.name }}: {{ attribute_type[attr] }}{{ ') -> None:' if loop.last else ',' }}
    {% endif %}{# /if not attr.required #}
    {% endfor %}{# /for attr #}
        """Initializes with the given values."""
        {% set newliner = joiner('SWITCH') %}
        {% for attr in classdef.attributes.values() %}
        {% if newliner() %}{{ '\\n' }}{% endif %}
        {% if attr.description %}
        {{ attr.description|comment|indent|indent }}
        {% endif %}
        self.{{ attr.name }} = {{ attr.name }}
        {% endfor %}

    def to_jsonable(self) -> MutableMapping[str, Any]:
        """
        Dispatches the conversion to {{ classdef.identifier|snake_case }}_to_jsonable.

        :return: JSON-able representation
        """
        return {{ classdef.identifier|snake_case }}_to_jsonable(self)''')


@icontract.require(lambda classdef: classdef.identifier != '', 'Anonymous classes not allowed', enabled=True)
@icontract.ensure(lambda result: not result.endswith('\n'))
def _generate_class_definition(classdef: Classdef) -> str:
    """
    Generate the Python code defining the class given the class definition.

    :param classdef: class definition in Python representation
    :return: Python code
    """
    if len(classdef.attributes) == 0:
        return _CLASS_DEF_WO_ATTRIBUTES.render(classdef=classdef)

    attribute_type = dict()
    for attr in classdef.attributes.values():
        if attr.typedef is None:
            raise ValueError('Unexpected None typedef of attr {!r} in class {!r}'.format(
                attr.name, classdef.identifier))

        attribute_type[attr] = _type_expression(
            typedef=attr.typedef, path='{}.{}'.format(classdef.identifier, attr.name))

    return _CLASS_DEF_WITH_ATTRIBUTES_TPL.render(classdef=classdef, attribute_type=attribute_type)


def _default_value(typedef: Typedef) -> str:
    """
    Determine the default value of the given type definition.

    :param typedef: type definition in Python representation.
    :return: Python code
    """
    # pylint: disable=too-many-return-statements
    if isinstance(typedef, Booldef):
        return 'False'
    elif isinstance(typedef, Intdef):
        return '0'
    elif isinstance(typedef, Floatdef):
        return '0.0'
    elif isinstance(typedef, Strdef):
        return "''"
    elif isinstance(typedef, Bytesdef):
        return "b''"
    elif isinstance(typedef, Listdef):
        return '[]'
    elif isinstance(typedef, Dictdef):
        return 'dict()'
    elif isinstance(typedef, Classdef):
        return 'new_{}()'.format(swagger_to.snake_case(identifier=typedef.identifier))
    else:
        raise NotImplementedError('Translating the typedef to a default value is not supported: {}'.format(typedef))


_FACTORY_METHOD_TPL = _from_string_with_informative_exceptions(
    env=_ENV,
    text='''\
def new_{{ classdef.identifier|snake_case }}() -> {{ classdef.identifier }}:
    """Generates an instance of {{ classdef.identifier }} with default values."""
    {% if not required_attributes %}
    return {{ classdef.identifier }}()
    {% else %}
    return {{ classdef.identifier }}(
        {% for attr in required_attributes %}
        {{ attr.name }}={{ default_value[attr] }}{{ ')' if loop.last else ',' }}
        {% endfor %}
    {% endif %}{# /if not required_attrbiutes #}
''')


@icontract.ensure(lambda result: not result.endswith('\n'))
def _generate_factory_method(classdef: Classdef) -> str:
    """
    Generate the code of the factory method for a class conforming to the given class definition.

    :param classdef: class definition in Python representation
    :return: Python code
    """
    default_value = dict()
    for attr in classdef.attributes.values():
        if attr.typedef is None:
            raise ValueError('Unexpected None typedef of attr {!r}'.format(attr.name))

        default_value[attr] = _default_value(typedef=attr.typedef)

    return _FACTORY_METHOD_TPL.render(
        classdef=classdef,
        required_attributes=[attr for attr in classdef.attributes.values() if attr.required],
        default_value=default_value).strip()


_FROM_OBJ_TPL = _from_string_with_informative_exceptions(
    env=_ENV,
    text='''\
def from_obj(obj: Any, expected: List[type], path: str = '') -> Any:
    """
    Checks and converts the given obj along the expected types.

    :param obj: to be converted
    :param expected: list of types representing the (nested) structure
    :param path: to the object used for debugging
    :return: the converted object
    """
    if not expected:
        raise ValueError("`expected` is empty, but at least one type needs to be specified.")

    exp = expected[0]

    if exp == float:
        if isinstance(obj, int):
            return float(obj)

        if isinstance(obj, float):
            return obj

        raise ValueError(
            'Expected object of type int or float at {!r}, but got {}.'.format(path, type(obj)))

    if exp in [bool, int, str, list, dict]:
        if not isinstance(obj, exp):
            raise ValueError(
                'Expected object of type {} at {!r}, but got {}.'.format(exp, path, type(obj)))

    if exp in [bool, int, float, str]:
        return obj

    if exp == list:
        lst = []  # type: List[Any]
        for i, value in enumerate(obj):
            lst.append(
                from_obj(value, expected=expected[1:], path='{}[{}]'.format(path, i)))

        return lst

    if exp == dict:
        adict = dict()  # type: Dict[str, Any]
        for key, value in obj.items():
            if not isinstance(key, str):
                raise ValueError(
                    'Expected a key of type str at path {!r}, got: {}'.format(path, type(key)))

            adict[key] = from_obj(value, expected=expected[1:], path='{}[{!r}]'.format(path, key))

        return adict
    {% for classdef in classdefs %}

    if exp == {{ classdef.identifier }}:
        return {{ classdef.identifier|snake_case }}_from_obj(obj, path=path)
    {% endfor %}

    raise ValueError("Unexpected `expected` type: {}".format(exp))''')


@icontract.ensure(lambda result: not result.endswith('\n'))
def _generate_from_obj(classdefs: List[Classdef]) -> str:
    """
    Generate the code of the ``from_obj`` function.

    :param classdefs: all available class definitions in Python representation
    :return: Python code
    """
    return _FROM_OBJ_TPL.render(classdefs=classdefs)


def _expected_type_expression(typedef: Typedef) -> str:
    """
    Determine the type expression supplied to ``from_obj`` function corresponding to the type definition.

    :param typedef: type definition in Python representation
    :return: Python code representing the type definition
    """
    # pylint: disable=too-many-return-statements
    if isinstance(typedef, Booldef):
        return 'bool'
    elif isinstance(typedef, Intdef):
        return 'int'
    elif isinstance(typedef, Floatdef):
        return 'float'
    elif isinstance(typedef, Strdef):
        return 'str'
    elif isinstance(typedef, Bytesdef):
        return 'bytes'
    elif isinstance(typedef, Listdef):
        if typedef.items is None:
            raise ValueError('Unexpected None items in typedef: {!r}'.format(typedef.identifier))

        return 'list, {}'.format(_expected_type_expression(typedef=typedef.items))
    elif isinstance(typedef, Dictdef):
        if typedef.values is None:
            raise ValueError('Unexpected None values in typedef: {!r}'.format(typedef.identifier))

        return 'dict, {}'.format(_expected_type_expression(typedef=typedef.values))
    elif isinstance(typedef, Classdef):
        return typedef.identifier
    else:
        raise NotImplementedError('Translating the typedef to an expected type is not supported: {}'.format(typedef))


_CLASS_FROM_OBJ_TPL = _from_string_with_informative_exceptions(
    env=_ENV,
    text='''\
def {{ classdef.identifier|snake_case }}_from_obj(obj: Any, path: str = "") -> {{ classdef.identifier }}:
    """
    Generates an instance of {{ classdef.identifier }} from a dictionary object.

    :param obj: a JSON-ed dictionary object representing an instance of {{ classdef.identifier }}
    :param path: path to the object used for debugging
    :return: parsed instance of {{ classdef.identifier }}
    """
    if not isinstance(obj, dict):
        raise ValueError('Expected a dict at path {}, but got: {}'.format(path, type(obj)))

    {% if not classdef.attributes %}
    return {{ classdef.identifier }}()
    {% else %}
    for key in obj:
        if not isinstance(key, str):
            raise ValueError(
                'Expected a key of type str at path {}, but got: {}'.format(path, type(key)))
    {% for attr in classdef.attributes.values() %}

    {% if attr.required %}
    {{ attr.name }}_from_obj = from_obj(
        obj[{{ attr.name|repr }}],
        expected=[{{ expected_type_expression[attr] }}],
        path=path + {{ '.%s'|format(attr.name)|repr }})  # type: {{ type_expression[attr] }}
    {% else %}
    if {{ attr.name|repr }} in obj:
        {{ attr.name }}_from_obj = from_obj(
            obj[{{ attr.name|repr }}],
            expected=[{{ expected_type_expression[attr] }}],
            path=path + {{ '.%s'|format(attr.name)|repr }})  # type: Optional[{{ type_expression[attr] }}]
    else:
        {{ attr.name }}_from_obj = None
    {% endif %}{# /if attr.required #}
    {% endfor %}{# /for attr in classdef.attributes.values() #}

    return {{ classdef.identifier }}(
    {% for attr in classdef.attributes.values() %}
        {{ attr.name }}={{ attr.name }}_from_obj{{ ')' if loop.last else ',' }}
    {% endfor %}{# /for attr in classdef.attributes.values() #}
    {% endif %}{# /if not classdef.attributes #}''')


@icontract.ensure(lambda result: not result.endswith('\n'))
def _generate_class_from_obj(classdef: Classdef) -> str:
    """
    Generate the code of the ``{class}_from_obj`` function that parses the JSON-ed object to an instance of a class.

    :param classdef: class definition in Python representation
    :return: Python code
    """
    expected_type_expression = dict()
    type_expression = dict()
    for attr in classdef.attributes.values():
        if attr.typedef is None:
            raise ValueError('Unexpected None typedef of attr {!r} in class {!r}'.format(
                attr.name, classdef.identifier))

        expected_type_expression[attr] = _expected_type_expression(typedef=attr.typedef)
        type_expression[attr] = _type_expression(typedef=attr.typedef, path=classdef.identifier + '.' + attr.name)

    return _CLASS_FROM_OBJ_TPL.render(
        classdef=classdef, expected_type_expression=expected_type_expression, type_expression=type_expression).strip()


_TO_JSONABLE_TPL = _from_string_with_informative_exceptions(
    env=_ENV,
    text='''\
def to_jsonable(obj: Any, expected: List[type], path: str = "") -> Any:
    """
    Checks and converts the given object along the expected types to a JSON-able representation.

    :param obj: to be converted
    :param expected: list of types representing the (nested) structure
    :param path: path to the object used for debugging
    :return: JSON-able representation of the object
    """
    if not expected:
        raise ValueError("`expected` is empty, but at least one type needs to be specified.")

    exp = expected[0]
    if not isinstance(obj, exp):
        raise ValueError('Expected object of type {} at path {!r}, but got {}.'.format(
            exp, path, type(obj)))

    # Assert on primitive types to help type-hinting.
    if exp == bool:
        assert isinstance(obj, bool)
        return obj

    if exp == int:
        assert isinstance(obj, int)
        return obj

    if exp == float:
        assert isinstance(obj, float)
        return obj

    if exp == str:
        assert isinstance(obj, str)
        return obj

    if exp == list:
        assert isinstance(obj, list)

        lst = []  # type: List[Any]
        for i, value in enumerate(obj):
            lst.append(
                to_jsonable(value, expected=expected[1:], path='{}[{}]'.format(path, i)))

        return lst

    if exp == dict:
        assert isinstance(obj, dict)

        adict = dict()  # type: Dict[str, Any]
        for key, value in obj.items():
            if not isinstance(key, str):
                raise ValueError(
                    'Expected a key of type str at path {!r}, got: {}'.format(path, type(key)))

            adict[key] = to_jsonable(
                value,
                expected=expected[1:],
                path='{}[{!r}]'.format(path, key))

        return adict
    {% for classdef in classdefs %}

    if exp == {{ classdef.identifier }}:
        assert isinstance(obj, {{ classdef.identifier }})
        return {{ classdef.identifier|snake_case }}_to_jsonable(obj, path=path)
    {% endfor %}{# /for classdef in classdefs #}

    raise ValueError("Unexpected `expected` type: {}".format(exp))''')


@icontract.ensure(lambda result: not result.endswith('\n'))
def _generate_to_jsonable(classdefs: List[Classdef]) -> str:
    """
    Generate the code of the ``to_jsonable`` function that converts a Python object to a JSON-able format.

    :param classdefs: list of all available class definitions
    :return: Python code
    """
    return _TO_JSONABLE_TPL.render(classdefs=classdefs)


_CLASS_TO_JSONABLE_TPL = _from_string_with_informative_exceptions(
    env=_ENV,
    text='''\
def {{ classdef.identifier|snake_case }}_to_jsonable(
        {{ classdef.identifier|snake_case }}: {{ classdef.identifier }},
        path: str = "") -> MutableMapping[str, Any]:
    """
    Generates a JSON-able mapping from an instance of {{ classdef.identifier }}.

    :param {{ classdef.identifier|snake_case }}: instance of {{ classdef.identifier }} to be JSON-ized
    :param path: path to the {{ classdef.identifier|snake_case }} used for debugging
    :return: a JSON-able representation
    """
    {% if not classdef.attributes %}
    return dict()
    {% else %}
    res = dict()  # type: Dict[str, Any]
    {% for attr in classdef.attributes.values() %}

    {% set assignment %}
{% if is_primitive[attr] %}
res[{{ attr.name|repr }}] = {{ classdef.identifier|snake_case }}.{{ attr.name }}
{% else %}
res[{{ attr.name|repr }}] = to_jsonable(
    {{ classdef.identifier|snake_case }}.{{ attr.name }},
    expected=[{{ expected_type_expression[attr] }}],
    path={{ '{}.%s'|format(attr.name)|repr }}.format(path))
{% endif %}{# /if is_primitive[attr] #}
    {% endset %}{# /set assignment #}
    {% if not attr.required %}
    if {{ classdef.identifier|snake_case }}.{{ attr.name }} is not None:
        {{ assignment|trim|indent }}
    {% else %}
    {{ assignment|trim|indent }}
    {% endif %}{# /if not attr.required #}
    {% endfor %}{# /for attr in classdef.attributes.values() #}

    return res
    {% endif %}{# /if not classdef.attributes #}''')


@icontract.ensure(lambda result: not result.endswith('\n'))
def _generate_class_to_jsonable(classdef: Classdef) -> str:
    """
    Generate ``{class}_to_jsonable`` function which converts the given instance of the class to a JSON-able format.

    :param classdef: class definition in Python representation
    :return: Python code
    """
    is_primitive = dict()
    expected_type_expression = dict()

    for attr in classdef.attributes.values():
        if attr.typedef is None:
            raise ValueError('Unexpected None typedef of attr {!r} in class {!r}'.format(
                attr.name, classdef.identifier))

        is_primitive[attr] = isinstance(attr.typedef, (Booldef, Intdef, Floatdef, Strdef))
        expected_type_expression[attr] = _expected_type_expression(typedef=attr.typedef)

    return _CLASS_TO_JSONABLE_TPL.render(
        classdef=classdef, is_primitive=is_primitive, expected_type_expression=expected_type_expression).strip()


_REQUEST_DOCSTRING_TPL = _from_string_with_informative_exceptions(
    env=_ENV,
    text='''\
{% if request.description %}
{{ request.description|upper_first }}
{% else %}
Send a {{ request.method }} request to {{ request.path }}.
{% endif %}{# /if request.description #}
{% if request.parameters %}

{% for param in request.parameters %}
{% if not param.description %}
:param {{ param.identifier }}:
{% else %}
{% if '\\n' in param.description %}
:param {{ param.identifier }}:
    {{ param.description|indent }}
{% else %}
:param {{ param.identifier }}: {{ param.description }}
{% endif %}{# /if '\\n' in param.description #}
{% endif %}{# /if not param.description #}
{% endfor %}{# /for request.parameters #}
{% endif %}{# /if request.parameters #}
{% if resp is none or resp.description == ''%}

:return:
{% else %}
{% if '\\n' in resp.description %}

:return:
    {{ resp.description|indent }}
{% else %}

:return: {{ resp.description }}
{% endif %}{# /if '\\n' in resp.description #}
{% endif %}{# /if resp is none #}''')

_REQUEST_FUNCTION_TPL = _from_string_with_informative_exceptions(
    env=_ENV,
    text='''\
{% if not request.parameters %}
def {{ request.operation_id}}(self) -> {{ return_type }}:
{% else %}
{% set suffix = ') -> %s:'|format(return_type) %}
def {{ request.operation_id}}(
        self,
        {% for param in request.parameters %}
        {% if not param.required %}
        {{ param.identifier }}: Optional[{{ type_expression[param] }}] = None{{ suffix if loop.last else ',' }}
        {% else %}
        {{ param.identifier }}: {{ type_expression[param] }}{{ suffix if loop.last else ',' }}
        {% endif %}
        {% endfor %}{# /for param in request.parameters #}
{% endif %}{# /if not request.parameters #}
    {{ request_docstring|docstring|indent }}
    {% if not path_tokens %}{### Path parameters ###}
    url = self.url_prefix + {{ request.path|repr }}
    {% else %}
    url = "".join([
        self.url_prefix,
    {% for token in path_tokens %}
    {% if token.parameter %}
        {{ 'str(%s)'|format(token.parameter.identifier if is_str[token.parameter] else token.parameter.identifier) }}{#
            #}{{ '])' if loop.last else ',' }}
    {% else %}
        {{ token.text|repr }}{{ '])' if loop.last else ',' }}
    {% endif %}
    {% endfor %}{# /for token in path_tokens #}
    {% endif %}{# /if not path_tokens #}
    {% if request.header_parameters %}{### Header parameters ###}

    headers = {}  # type: Dict[str, str]
        {% for param in request.header_parameters %}

            {% set set_header_item %}
                {% if is_str[param] %}
headers[{{ param.name|repr }}] = {{ param.identifier }}
                {% elif is_primitive[param] %}
headers[{{ param.name|repr }}] = json.dumps({{ param.identifier }})
                {% else %}
headers[{{ param.name|repr }}] = json.dumps(
    to_jsonable(
        {{ param.identifier }},
        expected=[{{ expected_type_expression[param] }}]))
                {% endif %}{# /if is_primitive[param] #}
            {% endset %}
            {% if param.required %}
    {{ set_header_item|trim|indent }}
            {% else %}
    if {{ param.identifier }} is not None:
        {{ set_header_item|trim|indent|indent }}
            {% endif %}{# /if param.required #}
        {% endfor %}{# /for param in request.header_parameters #}
    {% endif %}{# /if request.header_parameters #}
    {% if request.query_parameters %}{### Query parameters ###}

    params = {}  # type: Dict[str, str]
        {% for param in request.query_parameters %}

            {% set set_params_item %}
                {% if is_str[param] %}
params[{{ param.name|repr }}] = {{ param.identifier }}
                {% elif is_primitive[param] %}
params[{{ param.name|repr }}] = json.dumps({{ param.identifier }})
                {% else %}
params[{{ param.name|repr }}] = json.dumps(
    to_jsonable(
        {{ param.identifier }},
        expected=[{{ expected_type_expression[param] }}]))
                {% endif %}{# /if is_primitive[param] #}
            {% endset %}
            {% if param.required %}
    {{ set_params_item|trim|indent }}
            {% else %}
    if {{ param.identifier }} is not None:
        {{ set_params_item|trim|indent|indent }}
            {% endif %}{# /if param.required #}
        {% endfor %}{# /for param in request.query_parameters #}
    {% endif %}{# /if request.query_parameters #}
    {% if request.body_parameter %}{### Body parameter ###}

    {% set set_body %}
        {% if is_primitive[request.body_parameter] %}
data = {{ request.body_parameter.identifier }}
        {% else %}
data = to_jsonable(
    {{ request.body_parameter.identifier }},
    expected=[{{ expected_type_expression[request.body_parameter] }}])
        {% endif %}{# /is_primitive[request.body_parameter] #}
    {% endset %}
    {% if request.body_parameter.required %}
    {{ set_body|indent }}
    {% else %}
    data = None  # type: Optional[Any]
    if {{ request.body_parameter.identifier }} != None:
        {{ set_body|trim|indent|indent }}
    {% endif %}{# /if request.body_parameter.required #}
    {% endif %}{# /if request.body_parameter #}
    {% if request.formdata_parameters %}{### Form-data parameters ###}

    data = {}  # type: Dict[str, str]
    {% for param in request.formdata_parameters %}

        {% set set_data_item %}
            {% if is_str[param] %}
data[{{ param.name|repr }}] = {{ param.identifier }}
            {% elif is_primitive[param] %}
data[{{ param.name|repr }}] = json.dumps({{ param.identifier }})
            {% else %}
data[{{ param.name|repr }}] = json.dumps(
    to_jsonable(
        {{ param.identifier}},
        expected=[{{ expected_type_expression[param] }}]))
            {% endif %}{# /if is_primitive[param] #}
        {% endset %}
        {% if param.required %}
    {{ set_data_item|trim|indent|indent }}
        {% else %}
    if {{ param.identifier }} is not None:
        {{ set_data_item|trim|indent|indent }}
        {% endif %}{# /if param.required #}
    {% endfor %}{# /for param in request.formdata_parameters #}
    {% endif %}{# /if request.formdata_parameters #}
    {% if request.file_parameters %}{### File parameters ###}

    files = {}  # type: Dict[str, BinaryIO]
        {% for param in request.file_parameters %}

            {% if param.required %}
    files[{{ param.name|repr }}] = {{ param.identifier }}
            {% else %}
    if {{ param.identifier }} is not None:
        files[{{ param.name|repr }}] = {{ param.identifier }}
            {% endif %}{# /if param.required #}
        {% endfor %}{# /for param in request.file_parameters #}
    {% endif %}{# /if request.file_parameters #}

    {% if not request.parameters %}
    resp = requests.request(method={{ request.method|repr }}, url=url, auth=self.auth)
    {% else %}
    resp = requests.request(
        method={{ request.method|repr }},
        url=url,
        {% if request.header_parameters %}
        headers=headers,
        {% endif %}
        {% if request.query_parameters %}
        params=params,
        {% endif %}
        {% if request.body_parameter %}
        json=data,
        {% endif %}
        {% if request.formdata_parameters %}
        data=data,
        {% endif %}
        {% if request.file_parameters %}
        files=files,
        {% endif %}
        auth=self.auth)
    {% endif %}{# /if not request.parameters #}

    with contextlib.closing(resp):
        resp.raise_for_status()
        {% if return_type == 'bytes' %}
        return resp.content
        {% elif return_type == 'MutableMapping[str, Any]' %}
        return resp.json()
        {% else %}
        return from_obj(
            obj=resp.json(),
            expected=[{{ expected_type_expression[resp] }}])
        {% endif %}''')


class _Token:
    """Represent a token of the tokenized Swagger path to an endpoint."""

    def __init__(self, text: str, parameter: Optional[Parameter] = None) -> None:
        """
        Initialize with the given values.

        :param text: text encompassed by the token
        :param parameter: parameter, if referenced by the token
        """
        self.text = text
        self.parameter = parameter


@icontract.require(
    lambda request: not request.body_parameter or not request.formdata_parameters,
    'Both body parameter and form-data parameters are specified. '
    'The python client does not know how to resolve this request.',
    enabled=True)
@icontract.ensure(lambda result: not result.endswith('\n'))
def _generate_request_function(request: Request) -> str:
    """
    Generate the code of the client request function.

    :param request: request to the endpoint in Python representation
    :return: Python code
    """
    ##
    # Prepare response
    ##

    resp = None  # type: Optional[Response]
    return_type = 'bytes'

    if '200' in request.responses:
        resp = request.responses['200']

        if request.produces == ['application/json']:
            if resp.typedef is not None:
                return_type = _type_expression(typedef=resp.typedef, path=request.operation_id + '.' + str(resp.code))
            else:
                # The schema for the response has not been defined. Hence we can not parse the response to an object,
                # but we can at least parse it as JSON.
                return_type = 'MutableMapping[str, Any]'

    ##
    # Preapre request docstring
    ##

    request_docstring = _REQUEST_DOCSTRING_TPL.render(request=request, resp=resp).rstrip()

    ##
    # Prepare a representation of path parameters
    ##

    token_pth = swagger_to.tokenize_path(path=request.path)
    name_to_parameters = {param.name: param for param in request.parameters}

    path_tokens = []  # type: List[_Token]
    if token_pth.parameter_to_token_indices:
        for i, token_text in enumerate(token_pth.tokens):
            param = None  # type: Optional[Parameter]
            if i in token_pth.token_index_to_parameter:
                param_name = token_pth.token_index_to_parameter[i]
                param = name_to_parameters[param_name]

            path_tokens.append(_Token(text=token_text, parameter=param))

    ##
    # Prepare expected types
    ##

    expected_type_expression = dict()  # type: Dict[Union[Parameter, Response], str]
    for param in request.parameters:
        if isinstance(param.typedef, Filedef):
            continue

        if param.typedef is None:
            raise ValueError('Unexpected None typedef in param {!r} of request {!r}'.format(
                param.name, request.operation_id))

        expected_type_expression[param] = _expected_type_expression(typedef=param.typedef)

    if return_type not in ['bytes', 'MutableMapping[str, Any]']:
        if resp is None:
            raise ValueError('Unexpected None resp with return_type {!r} in request {!r}'.format(
                return_type, request.operation_id))

        if resp.typedef is None:
            raise ValueError('Unexpected None resp.typedef with return_type {!r} in request {!r}'.format(
                return_type, request.operation_id))

        expected_type_expression[resp] = _expected_type_expression(typedef=resp.typedef)

    ##
    # Render
    ##

    type_expression = dict()
    for param in request.parameters:
        if param.typedef is None:
            raise ValueError('Unexpected None typedef of param {!r} in request {!r}'.format(
                param.name, request.operation_id))

        type_expression[param] = _type_expression(
            typedef=param.typedef, path='{}.{}'.format(request.operation_id, param.name))

    return _REQUEST_FUNCTION_TPL.render(
        request=request,
        return_type=return_type,
        resp=resp,
        request_docstring=request_docstring,
        type_expression=type_expression,
        path_tokens=path_tokens,
        is_str={param: isinstance(param.typedef, Strdef)
                for param in request.parameters},
        is_primitive={
            param: isinstance(param.typedef, (Booldef, Intdef, Floatdef, Strdef))
            for param in request.parameters
        },
        expected_type_expression=expected_type_expression).strip()


_CLIENT_PY = _from_string_with_informative_exceptions(
    env=_ENV,
    text='''\
#!/usr/bin/env python3
# Automatically generated file by swagger_to. DO NOT EDIT OR APPEND ANYTHING!
"""Implements the client for {{ service_name }}."""

# pylint: skip-file
# pydocstyle: add-ignore=D105,D107,D401

import contextlib
import json
from typing import Any, BinaryIO, Dict, List, MutableMapping, Optional

import requests
import requests.auth
{% if classdefs %}


{{ from_obj }}


{{ to_jsonable }}
{% for classdef in classdefs %}


{{ class_definition[classdef] }}


{{ factory_method[classdef] }}


{{ class_from_obj[classdef] }}


{{ class_to_jsonable[classdef] }}
{% endfor %}{# /for classdef in classdefs #}
{% endif %}{# /if classdefs #}


class RemoteCaller:
    """Executes the remote calls to the server."""

    def __init__(self, url_prefix: str, auth: Optional[requests.auth.AuthBase] = None) -> None:
        self.url_prefix = url_prefix
        self.auth = auth
    {% for request in requests %}

    {{ request_function[request]|indent }}
    {% endfor %}{# /for request in requests #}


# Automatically generated file by swagger_to. DO NOT EDIT OR APPEND ANYTHING!

''')


@icontract.ensure(lambda result: result.endswith('\n'), 'File ends with a new line.')
def generate_client_py(service_name: str, typedefs: MutableMapping[str, Typedef], requests: List[Request]) -> str:
    """
    Generate the client code.

    :param service_name: used to designate the service that client connects to
    :param typedefs: table of type definitions in Python representation
    :param requests: request functions in Python representation
    :return: Python code
    """
    classdefs = [typedef for typedef in typedefs.values() if isinstance(typedef, Classdef)]

    assert len(set(classdefs)) == len(classdefs), \
        'All class definitions in Python representation are expected to be unique.'

    return _CLIENT_PY.render(
        service_name=service_name,
        classdefs=classdefs,
        from_obj=_generate_from_obj(classdefs=classdefs),
        to_jsonable=_generate_to_jsonable(classdefs=classdefs),
        class_definition={classdef: _generate_class_definition(classdef=classdef)
                          for classdef in classdefs},
        factory_method={classdef: _generate_factory_method(classdef=classdef)
                        for classdef in classdefs},
        class_from_obj={classdef: _generate_class_from_obj(classdef=classdef)
                        for classdef in classdefs},
        class_to_jsonable={classdef: _generate_class_to_jsonable(classdef=classdef)
                           for classdef in classdefs},
        requests=requests,
        request_function={request: _generate_request_function(request=request)
                          for request in requests})



"""Parse Swagger specification and generates server and client stubs."""
import re
import string
from typing import List, MutableMapping, Tuple

import collections
import icontract

# pylint: disable=missing-docstring

VARIABLE_RE = re.compile(r'^[a-zA-Z0-9_]+$')


@icontract.ensure(lambda result: VARIABLE_RE.match(result))
def parse_definition_ref(ref: str) -> str:
    """Parse a reference to a definition and return the definition name."""
    prefix = '#/definitions/'
    if not ref.startswith(prefix):
        raise ValueError("Expected a ref with prefix {!r}, but got: {!r}".format(prefix, ref))

    return ref[len(prefix):]


@icontract.ensure(lambda result: VARIABLE_RE.match(result))
def parse_parameter_ref(ref: str) -> str:
    """Parse a reference to a parameter and return the parameter name."""
    prefix = '#/parameters/'
    if not ref.startswith(prefix):
        raise ValueError("Expected a ref with prefix {!r}, but got: {!r}".format(prefix, ref))

    return ref[len(prefix):]


# Abbreviations to be treated specially in snake_case and camelCase conversions
SPECIALS = ['URLs', 'IDs', 'URL', 'ID', 'HTTP', 'HTTPS']


def camel_case_split(identifier: str) -> List[str]:
    """
    Split the identifier given in camel case into parts.

    >>> camel_case_split(identifier='CamelCase')
    ['Camel', 'Case']

    >>> camel_case_split(identifier='CamelURLs')
    ['Camel', 'URLs']
    """
    if identifier == '':
        raise ValueError("Unexpected empty identifier")

    cur = 0

    parts = []  # type: List[str]
    while cur < len(identifier):
        found_special = False
        for special in SPECIALS:
            if identifier[cur:cur + len(special)] == special:
                parts.append(special)
                cur += len(special)
                found_special = True

        if not found_special:
            if identifier[cur] in string.ascii_uppercase:
                parts.append('')
                parts[-1] += identifier[cur]
                cur += 1
            else:
                if len(parts) == 0:
                    parts.append('')

                parts[-1] += identifier[cur]
                cur += 1

    return parts


_NONLETTER_PREFIX_RE = re.compile(r'^[^a-zA-Z]+')
_NONLETTER_SUFFIX_RE = re.compile(r'[^a-zA-Z]+$')


def _trim_nonletters(identifier: str) -> Tuple[str, str, str]:
    """
    Trim the non-letter prefix and suffix from the identifier.

    >>> _trim_nonletters('_1some_text23_')
    ('_1', 'some_text', '23_')

    >>> _trim_nonletters('_1some_text')
    ('_1', 'some_text', '')

    >>> _trim_nonletters('some_text23_')
    ('', 'some_text', '23_')

    >>> _trim_nonletters('')
    ('', '', '')

    :param identifier: to be converted
    :return: lowercase snake_case identifier
    """
    if not identifier:
        return '', identifier, ''

    prefix = ''
    prefix_mtch = _NONLETTER_PREFIX_RE.search(identifier)
    if prefix_mtch:
        prefix = prefix_mtch.group(0)

    suffix = ''
    suffix_mtch = _NONLETTER_SUFFIX_RE.search(identifier)
    if suffix_mtch:
        suffix = suffix_mtch.group(0)

    trimmed = identifier[len(prefix):len(identifier) - len(suffix)]

    return prefix, trimmed, suffix


@icontract.require(
    lambda identifier: identifier != '', error=lambda: ValueError("Unexpected empty identifier"), enabled=True)
@icontract.ensure(lambda result: '-' not in result)
@icontract.ensure(lambda result: result[0] == result[0].upper())
@icontract.ensure(lambda result: result != '')
def capital_camel_case(identifier: str) -> str:
    """
    Translate an arbitrary identifier to a CamelCase.

    >>> capital_camel_case(identifier='camelCase')
    'CamelCase'

    >>> capital_camel_case(identifier='snake_case')
    'SnakeCase'

    >>> capital_camel_case(identifier='Dash-Case')
    'DashCase'

    >>> capital_camel_case(identifier='dash-case')
    'DashCase'

    >>> capital_camel_case(identifier='_snake_case')
    '_SnakeCase'

    >>> capital_camel_case(identifier='snake_case_')
    'SnakeCase_'

    >>> capital_camel_case(identifier='__')
    '__'

    :param identifier: arbitrary identifier
    :return: identifier as CamelCase
    """
    prefix, trimmed, suffix = _trim_nonletters(identifier=identifier)

    if not trimmed:
        return identifier

    # yapf: disable
    parts = [
        part
        for underscore_part in trimmed.split("_")
        for dash_part in underscore_part.split("-")
        for part in camel_case_split(identifier=dash_part)
    ]
    # yapf: enable

    new_parts = []  # type: List[str]
    for part in parts:
        part = part.lower()

        if part in ['url', 'id']:
            new_parts.append(part.upper())
        elif part == 'urls':
            new_parts.append('URLs')
        elif part == 'ids':
            new_parts.append('IDs')
        else:
            new_parts.append(part[0].upper() + part[1:].lower())

    return "".join([prefix] + new_parts + [suffix])


@icontract.require(
    lambda identifier: identifier != '', error=lambda: ValueError("Unexpected empty identifier"), enabled=True)
@icontract.ensure(lambda result: '-' not in result)
@icontract.ensure(lambda result: result[0] == result[0].lower())
@icontract.ensure(lambda result: result != '')
def camel_case(identifier: str) -> str:
    """
    Translate an arbitrary identifier to a camelCase.

    >>> camel_case(identifier='CamelCase')
    'camelCase'

    >>> camel_case(identifier='snake_case')
    'snakeCase'

    >>> camel_case(identifier='Snake_case')
    'snakeCase'

    >>> camel_case(identifier='Dash-Case')
    'dashCase'

    >>> camel_case(identifier='dash-case')
    'dashCase'

    >>> camel_case(identifier='_snake_case')
    '_snakeCase'

    >>> camel_case(identifier='snake_case_')
    'snakeCase_'

    >>> camel_case(identifier='__')
    '__'

    :param identifier: arbitrary identifier
    :return: identifier as camelCase
    """
    prefix, trimmed, suffix = _trim_nonletters(identifier=identifier)

    if not trimmed:
        return identifier

    # yapf: disable
    parts = [
        part
        for underscore_part in trimmed.split("_")
        for dash_part in underscore_part.split("-")
        for part in camel_case_split(identifier=dash_part)
    ]
    # yapf: enable

    new_parts = [parts[0].lower()]  # type: List[str]

    for part in parts[1:]:
        part = part.lower()

        if part in ['url', 'id']:
            new_parts.append(part.upper())
        elif part == 'urls':
            new_parts.append('URLs')
        elif part == 'ids':
            new_parts.append('IDs')
        else:
            new_parts.append(part[0].upper() + part[1:].lower())

    return "".join([prefix] + new_parts + [suffix])


@icontract.require(
    lambda identifier: identifier != '', error=lambda: ValueError("Unexpected empty identifier"), enabled=True)
@icontract.ensure(lambda result: '-' not in result)
@icontract.ensure(lambda result: result == result.lower())
def snake_case(identifier: str) -> str:
    """
    Convert an identifier to a lowercase snake case.

    >>> snake_case(identifier='CamelCase')
    'camel_case'

    >>> snake_case(identifier='camelCase')
    'camel_case'

    >>> snake_case(identifier='snake_case')
    'snake_case'

    >>> snake_case(identifier='Snake_case')
    'snake_case'

    >>> snake_case(identifier='Dash-Case')
    'dash_case'

    >>> snake_case(identifier='dash-case')
    'dash_case'

    >>> snake_case(identifier='_CamelCase')
    '_camel_case'

    >>> snake_case(identifier='CamelCase_')
    'camel_case_'

    >>> snake_case(identifier='__')
    '__'

    :param identifier: to be converted
    :return: lowercase snake_case identifier
    """
    prefix = ''
    prefix_mtch = _NONLETTER_PREFIX_RE.search(identifier)
    if prefix_mtch:
        prefix = prefix_mtch.group(0)

    suffix = ''
    suffix_mtch = _NONLETTER_SUFFIX_RE.search(identifier)
    if suffix_mtch:
        suffix = suffix_mtch.group(0)

    trimmed = identifier[len(prefix):len(identifier) - len(suffix)]

    if not trimmed:
        return identifier

    # yapf: disable
    parts = [
        part
        for underscore_part in trimmed.split("_")
        for dash_part in underscore_part.split("-")
        for part in camel_case_split(identifier=dash_part)
    ]
    # yapf: enable

    result = '_'.join(parts)
    return ''.join([prefix, result.lower(), suffix])


def upper_first(text: str) -> str:
    """
    Capitalizes the first letter of the text.

    >>> upper_first(text='some text')
    'Some text'

    >>> upper_first(text='Some text')
    'Some text'

    >>> upper_first(text='')
    ''

    >>> upper_first(text='_some text')
    '_some text'

    :param text: to be capitalized
    :return: text with the first letter capitalized
    """
    if len(text) == 0:
        return ''

    return text[0].upper() + text[1:]


class TokenizedPath:
    """Represent a tokenization of a Swagger path to an endpoint."""

    def __init__(self):
        """Initialize with defaults."""
        self.tokens = []  # type: List[str]
        self.parameter_to_token_indices = collections.OrderedDict()  # type: MutableMapping[str, List[int]]
        self.token_index_to_parameter = collections.OrderedDict()  # type: MutableMapping[int, str]


PATH_TOKENIZATION_RE = re.compile(r'\{(?P<name>[a-zA-Z0-9_]*)\}|[^{]+|.')


def tokenize_path(path: str) -> TokenizedPath:
    """
    Tokenize the path coming from a Swagger spec to a dictionary such that you can modify the path parameters easily.

    :param path: original path
    :return: tokenized path with the dictionary
    """
    token_pth = TokenizedPath()

    for mtch in PATH_TOKENIZATION_RE.finditer(path):
        token_pth.tokens.append(mtch.group(0))

        if mtch.group("name") is not None:
            name = mtch.group("name")
            if name not in token_pth.parameter_to_token_indices:
                token_pth.parameter_to_token_indices[name] = []

            token_pth.parameter_to_token_indices[name].append(len(token_pth.tokens) - 1)
            token_pth.token_index_to_parameter[len(token_pth.tokens) - 1] = name

    return token_pth



#!/usr/bin/env python3
"""Generate server stubs from Swagger specification in Go."""

# pylint: disable=missing-docstring,too-many-instance-attributes,too-many-locals,too-many-ancestors,too-many-branches
# pylint: disable=too-many-statements, too-many-lines

from typing import MutableMapping, Union, Set, List, Optional, Mapping, Iterable, Tuple  # pylint: disable=unused-import

import collections
import icontract
import jinja2

import swagger_to
import swagger_to.indent
import swagger_to.intermediate
import swagger_to.swagger


class JsonSchema:
    """Represent a JSON validation schema."""

    def __init__(self):
        """Initialize with default values."""
        self.identifier = ''
        self.text = ''


def _to_json_schema(intermediate_schema: swagger_to.intermediate.JsonSchema) -> JsonSchema:
    """
    Convert the intermediate schema to a representation that we can use to easily generate go code.

    :param intermediate_schema: intermediate representation of a JSON schema
    :return: representation suitable for generation of Go code
    """
    schema = JsonSchema()
    schema.identifier = intermediate_schema.identifier
    schema.text = intermediate_schema.text

    return schema


class Typedef:
    """Represent a type definition such that it's suitable for generation of Go code."""

    def __init__(self):
        """Initialize with default values."""
        self.identifier = ''
        self.description = ''
        self.json_schema = None  # type: Union[None, JsonSchema]


class Fielddef:
    """Represent a field definition of a struct suitable for generation of Go code."""

    def __init__(self):
        """Initialize with default values."""
        self.typedef = None  # type: Union[None, Typedef]
        self.description = ''
        self.json_name = ''
        self.name = ''


class Structdef(Typedef):
    """Represent a struct type."""

    def __init__(self):
        """Initialize with default values."""
        super().__init__()

        self.fields = collections.OrderedDict()  # type: MutableMapping[str, Fielddef]
        self.required = []  # type: List[str]


class Arraydef(Typedef):
    """Represent an array type."""

    def __init__(self):
        """Initialize with default values."""
        super().__init__()

        self.items = None  # type: Union[None, Typedef]


class Mapdef(Typedef):
    """Represent a map type."""

    def __init__(self):
        """Initialize with default values."""
        super().__init__()

        self.values = None  # type: Union[None, Typedef]


class Pointerdef(Typedef):
    """Represent a pointer type."""

    def __init__(self):
        """Initialize with default values."""
        super().__init__()

        self.pointed = None  # type: Union[None, Typedef]


class Primitivedef(Typedef):
    """Represent a primitive type."""

    def __init__(self):
        """Initialize with default values."""
        super().__init__()

        self.type = ''


def _to_typedef(intermediate_typedef: swagger_to.intermediate.Typedef) -> Typedef:
    """Convert intermediate type definition into a type definition suitable for Go code generation."""
    typedef = None  # type: Union[None, Typedef]

    if isinstance(intermediate_typedef, swagger_to.intermediate.Primitivedef):
        typedef = Primitivedef()

        if intermediate_typedef.type == 'string':
            if intermediate_typedef.format == 'date-time':
                typedef.type = 'time.Time'
            else:
                typedef.type = 'string'
        elif intermediate_typedef.type == 'number':
            if intermediate_typedef.format == '':
                typedef.type = 'float64'
            elif intermediate_typedef.format == 'float':
                typedef.type = 'float32'
            elif intermediate_typedef.format == 'double':
                typedef.type = 'float64'
            else:
                raise ValueError("Unexpected format {!r} for type {!r}".format(intermediate_typedef.format,
                                                                               intermediate_typedef.type))

        elif intermediate_typedef.type == 'integer':
            if intermediate_typedef.format == '':
                typedef.type = 'int'
            elif intermediate_typedef.format == 'int32':
                typedef.type = 'int32'
            elif intermediate_typedef.format == 'int64':
                typedef.type = 'int64'
            else:
                raise ValueError("Unexpected format {!r} for type {!r}".format(intermediate_typedef.format,
                                                                               intermediate_typedef.type))

        elif intermediate_typedef.type == 'boolean':
            typedef.type = 'bool'

        else:
            raise NotImplementedError(
                "Unhandled translation of a primitive intermediate type to Go with 'type': {!r}".format(
                    intermediate_typedef.type))

    elif isinstance(intermediate_typedef, swagger_to.intermediate.Arraydef):
        if intermediate_typedef.items is None:
            raise ValueError("Unexpected intermediate type definition of an array to have items None: {!r}".format(
                intermediate_typedef.identifier))

        typedef = Arraydef()
        typedef.items = _to_typedef(intermediate_typedef=intermediate_typedef.items)

    elif isinstance(intermediate_typedef, swagger_to.intermediate.Mapdef):
        typedef = Mapdef()
        typedef.values = _to_typedef(intermediate_typedef=intermediate_typedef.values)

    elif isinstance(intermediate_typedef, swagger_to.intermediate.Objectdef):
        typedef = Structdef()

        for propdef in intermediate_typedef.properties.values():
            field = Fielddef()

            field_typedef = _to_typedef(intermediate_typedef=propdef.typedef)
            if not propdef.name in intermediate_typedef.required and isinstance(field_typedef, Primitivedef):
                optional_field_typedef = Pointerdef()
                optional_field_typedef.pointed = field_typedef
                field_typedef = optional_field_typedef

            field.typedef = field_typedef
            field.description = propdef.description
            field.json_name = propdef.name
            field.name = swagger_to.capital_camel_case(identifier=propdef.name)

            typedef.fields[field.name] = field

            if propdef.name in intermediate_typedef.required:
                typedef.required.append(field.name)

    else:
        raise NotImplementedError("Unhandled translation of an intermediate type to Go: {!r}".format(
            type(intermediate_typedef)))

    assert typedef is not None

    if intermediate_typedef.identifier != '':
        typedef.identifier = swagger_to.capital_camel_case(identifier=intermediate_typedef.identifier)

    typedef.description = intermediate_typedef.description

    typedef.json_schema = _to_json_schema(intermediate_schema=intermediate_typedef.json_schema)

    return typedef


@icontract.ensure(lambda result: all(key == typedef.identifier for key, typedef in result.items()))
def to_typedefs(
        intermediate_typedefs: MutableMapping[str, swagger_to.intermediate.Typedef]) -> MutableMapping[str, Typedef]:
    """Convert a table of intermediate type representations to a table of type definitions for generation of Go code."""
    typedefs = collections.OrderedDict()  # type: MutableMapping[str, Typedef]

    for intermediate_typedef in intermediate_typedefs.values():
        assert intermediate_typedef is not None

        typedef = _to_typedef(intermediate_typedef=intermediate_typedef)
        typedefs[typedef.identifier] = typedef

    return typedefs


# yapf: disable
@icontract.ensure(
    lambda intermediate_typedef, result:
    intermediate_typedef.identifier == '' or result.identifier == intermediate_typedef.identifier)
@icontract.ensure(
    lambda intermediate_typedef, result:
    intermediate_typedef.identifier != '' or result.identifier == '')
# yapf: enable
def _anonymous_or_get_typedef(intermediate_typedef: swagger_to.intermediate.Typedef,
                              typedefs: MutableMapping[str, Typedef]) -> Typedef:
    """Create an anonymous type definition or retrieve the type definition from the existing definition table."""
    if intermediate_typedef.identifier != '':
        identifier = swagger_to.capital_camel_case(identifier=intermediate_typedef.identifier)

        if not identifier in typedefs:
            raise ValueError("Undefined Go type for intermediate typedef {!r}: {!r}".format(
                intermediate_typedef.identifier, identifier))

        return typedefs[identifier]

    return _to_typedef(intermediate_typedef=intermediate_typedef)


def _walk(typedef: Typedef, parent: Optional[Typedef] = None) -> Iterable[Tuple[Optional[Typedef], Typedef]]:
    """Walk the tree of nested type definitions as (nesting type definition, nested type definition)."""
    yield parent, typedef

    if isinstance(typedef, Primitivedef):
        pass

    elif isinstance(typedef, Pointerdef):
        if typedef.pointed is None:
            raise ValueError("Unexpected None pointed in typedef: {!r}".format(typedef.identifier))

        yield from _walk(typedef=typedef.pointed, parent=typedef)

    elif isinstance(typedef, Arraydef):
        if typedef.items is None:
            raise ValueError("Unexpected None items in typedef: {!r}".format(typedef.identifier))

        yield from _walk(typedef=typedef.items, parent=typedef)

    elif isinstance(typedef, Mapdef):
        if typedef.values is None:
            raise ValueError("Unexpected None values in typedef: {!r}".format(typedef.identifier))

        yield from _walk(typedef=typedef.values, parent=typedef)

    elif isinstance(typedef, Structdef):
        for fielddef in typedef.fields.values():
            if fielddef.typedef is None:
                raise ValueError("Unexpected None typedef in fielddef {!r} of type {!r}".format(
                    fielddef.name, typedef.identifier))

            yield from _walk(typedef=fielddef.typedef, parent=typedef)

    else:
        raise NotImplementedError("_walk for Go type definition of type: {}".format(type(typedef)))


class Argument:
    """Represent an argument of a handler implementation."""

    def __init__(self):
        """Initialize with default values."""
        self.typedef = None  # type: Union[None, Typedef]
        self.identifier = ''
        self.in_what = ''

        # Original name of the endpoint parameter
        self.parameter_name = ''

        self.required = False
        self.parsing_identifier = ''
        self.json_schema = None  # type: Optional[JsonSchema]


class Handler:
    """Represent a handler interface of an endpoint."""

    def __init__(self):
        """Initialize with default values."""
        self.identifier = ''
        self.arguments = []  # type: List[Argument]


class Wrapper:
    """Represent a wrapper that parses the arguments from a request and forwards them to the handler."""

    def __init__(self):
        """Initialize with default values."""
        self.identifier = ''
        self.handler = None  # type: Union[None, Handler]

        self.header_arguments = []  # type: List[Argument]
        self.query_arguments = []  # type: List[Argument]
        self.path_arguments = []  # type: List[Argument]
        self.body_argument = None  # type: Union[None, Argument]


class Route:
    """Represent a muxing route to an endpoint."""

    def __init__(self):
        """Initialize with default values."""
        self.description = ''
        self.path = ''
        self.method = ''
        self.wrapper = Wrapper()
        self.handler = Handler()


def _endpoint_to_route_path(endpoint: swagger_to.intermediate.Endpoint) -> str:
    """
    Convert an endpoint path to Gorrila Mux route path.

    :param endpoint: whose path we need to convert
    :return: Gorrila Mux route path
    """
    token_pth = swagger_to.tokenize_path(path=endpoint.path)

    tkns = token_pth.tokens[:]

    for param in endpoint.parameters:
        if param.in_what != 'path':
            continue

        if param.name not in token_pth.parameter_to_token_indices:
            continue

        if not isinstance(param.typedef, swagger_to.intermediate.Primitivedef):
            raise ValueError("Expected path parameter {!r} in the path {!r} to be primitive, but got: {!r}".format(
                param.name, endpoint.path, type(param.typedef)))

        assert isinstance(param.typedef, swagger_to.intermediate.Primitivedef)
        if param.typedef.pattern != '':
            param_in_route = '{' + param.name + ":" + param.typedef.pattern + "}"
        else:
            param_in_route = '{' + param.name + '}'

        for i in token_pth.parameter_to_token_indices[param.name]:
            tkns[i] = param_in_route

    return "".join(tkns)


def _to_route(endpoint: swagger_to.intermediate.Endpoint, typedefs: MutableMapping[str, Typedef]) -> Route:
    """
    Convert an intermediate representation of an endpoint to a muxing route of Go server stub.

    :param endpoint: intermediate representation of an endpoint
    :param typedefs: table of type definitions
    :return: converted route
    """
    route = Route()
    route.method = endpoint.method.lower()
    route.path = _endpoint_to_route_path(endpoint=endpoint)
    route.description = endpoint.description

    ##
    # Determine handable parameters
    ##
    handable_parameters = []  # type: List[swagger_to.intermediate.Parameter]

    for param in endpoint.parameters:
        # Assert that we can handle all the supplied parameters.
        if param.in_what == 'formData':
            # No code is generated for the parameters in the form data since there are so many edge cases
            # which we possibly can't cover.
            continue
        elif param.in_what in ['query', 'body', 'path', 'header']:
            handable_parameters.append(param)
        else:
            raise NotImplementedError(
                "Handling of parameters in {} is not implemented yet: endpoint {} {}, parameter {}.".format(
                    param.in_what, endpoint.path, endpoint.method, param.name))

    ##
    # Generate identifiers corresponding to the parameters.
    ##

    param_to_identifier = {param: swagger_to.camel_case(identifier=param.name) for param in handable_parameters}

    # Add the location as prefix if the argument identifiers overlap
    identifiers = list(param_to_identifier.values())
    needs_location_prefix = len(set(identifiers)) != len(identifiers)
    if needs_location_prefix:
        param_to_identifier = {
            param: swagger_to.camel_case(identifier="{}_{}".format(param.in_what, param.name))
            for param in endpoint.parameters
        }

        ##
        # Assert that there are no conflicts at this point
        ##

        by_identifier = collections.defaultdict(
            list)  # type: MutableMapping[str, List[swagger_to.intermediate.Parameter]]
        for param, identifier in param_to_identifier.items():
            by_identifier[identifier].append(param)

        # yapf: disable
        msgs = [
            "in the endpoint {} {} for the identifier {!r}: {}".format(
                endpoint.method.upper(), endpoint.path, identifier, ", ".join(
                    ["{} in {}".format(param.name, param.in_what) for param in params]))
            for identifier, params in by_identifier.items()
            if len(params) > 1
        ]
        # yapf: enable

        if len(msgs) > 0:
            raise ValueError("There are conflicting identifiers for parameters:\n{}".format("\n".join(msgs)))

    ##
    # Convert parameters to arguments
    ##

    assert all(param in param_to_identifier for param in handable_parameters), \
        "Expected all parameters to have a generated argument identifier."

    for param in handable_parameters:
        identifier = param_to_identifier[param]

        argument = Argument()
        argument.typedef = _anonymous_or_get_typedef(intermediate_typedef=param.typedef, typedefs=typedefs)
        argument.required = param.required

        if not param.required and isinstance(argument.typedef, Primitivedef):
            pointer_typedef = Pointerdef()
            pointer_typedef.identifier = argument.typedef.identifier
            pointer_typedef.description = argument.typedef.description
            pointer_typedef.pointed = argument.typedef

            argument.typedef = pointer_typedef

        argument.parameter_name = param.name
        argument.identifier = identifier
        argument.in_what = param.in_what
        argument.parsing_identifier = swagger_to.camel_case(identifier='a_' + identifier)

        if param.json_schema is not None:
            argument.json_schema = _to_json_schema(intermediate_schema=param.json_schema)

        if argument.in_what == 'header':
            route.wrapper.header_arguments.append(argument)

        elif argument.in_what == 'query':
            route.wrapper.query_arguments.append(argument)

        elif argument.in_what == 'body':
            route.wrapper.body_argument = argument

        elif argument.in_what == 'path':
            route.wrapper.path_arguments.append(argument)

        else:
            raise AssertionError("Unexpected argument given in: {}".format(argument.in_what))

        route.handler.arguments.append(argument)

    ##
    # Determine route attributes
    ##

    route.wrapper.identifier = swagger_to.capital_camel_case(identifier='wrap_' + endpoint.operation_id)
    route.wrapper.handler = route.handler
    route.handler.identifier = swagger_to.capital_camel_case(identifier=endpoint.operation_id)

    return route


def to_routes(endpoints: List[swagger_to.intermediate.Endpoint], typedefs: MutableMapping[str, Typedef]) -> List[Route]:
    """
    Convert the intermediate representation of endpoints to muxing routes of a Go server stub.

    :param endpoints: intermediate representation of endpoints
    :param typedefs: table of type definitions
    :return: muxing routes of a Go server stub
    """
    routes = []  # type: List[Route]
    for endpoint in endpoints:
        routes.append(_to_route(endpoint=endpoint, typedefs=typedefs))

    return routes


@icontract.ensure(lambda result: not result.endswith('\n'))
def _comment(text: str) -> str:
    r"""
    Genearates a (possibly multi-line) comment from the text.

    >>> cmt = _comment("  testme\n  \nagain\n")
    >>> assert cmt == '//   testme\n//\n// again\n//'

    :param text: of the comment
    :return: Go code
    """
    out = []  # type: List[str]
    lines = text.split('\n')
    for line in lines:
        rstripped = line.rstrip()
        if len(rstripped) > 0:
            out.append('// {}'.format(rstripped))
        else:
            out.append('//')

    return '\n'.join(out)


@icontract.ensure(lambda result: result.startswith('"'))
@icontract.ensure(lambda result: result.endswith('"'))
def _escaped_str(text: str) -> str:
    """Escape the text and returns it as a valid Golang string."""
    return '"{}"'.format(
        text.replace('\\', '\\\\').replace('"', '\\"').replace('\a', '\\a').replace('\f', '\\f').replace('\t', '\\t')
        .replace('\n', '\\n').replace('\r', '\\r').replace('\v', '\\v'))


# Jinja2 environment
ENV = jinja2.Environment(trim_blocks=True, lstrip_blocks=True, loader=jinja2.BaseLoader())
ENV.filters.update({
    'capital_camel_case': swagger_to.capital_camel_case,
    'comment': _comment,
    'escaped_str': _escaped_str
})

ENV.globals.update({'is_pointerdef': lambda typedef: isinstance(typedef, Pointerdef)})

_STRUCT_TPL = ENV.from_string('''\
struct {
{% for fielddef in typedef.fields.values() %}
{% if not loop.first %}


{% endif %}
{% if fielddef.description %}
    {{ fielddef.description|comment|indent }}
{% endif %}
{% set json_tags = fielddef.json_name %}
{% if fielddef.name not in typedef.required %}
{% set json_tags = json_tags + ',omitempty' %}
{% endif %}
    {{ fielddef.name }} {{ field_type[fielddef] }} `json:{{ json_tags|escaped_str }}`{% if loop.last %}}{% endif %}
{% endfor %}
''')


@icontract.ensure(lambda result: result == result.strip())
def _express_type(typedef: Typedef) -> str:
    """Express the type in Golang corresponding to the type definition."""
    if isinstance(typedef, Primitivedef):
        return typedef.type

    if isinstance(typedef, Pointerdef):
        if typedef.pointed is None:
            raise ValueError("Unexpected None pointed in typedef: {!r}".format(typedef.identifier))

        return "*{}".format(_express_or_identify_type(typedef.pointed))

    if isinstance(typedef, Arraydef):
        if typedef.items is None:
            raise ValueError("Unexpected None items in typedef: {!r}".format(typedef.identifier))

        return "[]{}".format(_express_or_identify_type(typedef.items))

    if isinstance(typedef, Mapdef):
        if typedef.values is None:
            raise ValueError("Unexpected None values in typedef: {!r}".format(typedef.identifier))

        return "map[string]{}".format(_express_or_identify_type(typedef.values))

    if isinstance(typedef, Structdef):
        if len(typedef.fields) == 0:
            return "struct {}"

        field_type = dict()
        for fielddef in typedef.fields.values():
            if fielddef.typedef is None:
                raise ValueError('Unexpected None typedef of fielddef {!r} in typedef {!r}'.format(
                    fielddef.name, typedef.identifier))

            field_type[fielddef] = _express_or_identify_type(fielddef.typedef)

        return _STRUCT_TPL.render(typedef=typedef, field_type=field_type).strip()

    else:
        raise NotImplementedError("No Go type writing defined for typedef of type: {!r}".format(type(typedef)))


@icontract.ensure(lambda result: result == result.strip())
def _express_or_identify_type(typedef: Typedef) -> str:
    """Give the type identifier or expresses the type if the typedef lacks an identifier."""
    if typedef.identifier != '':
        return typedef.identifier

    return _express_type(typedef=typedef)


@icontract.require(lambda typedef: typedef.identifier != '')
@icontract.ensure(lambda result: result == result.strip())
def _define_type(typedef: Typedef) -> str:
    """Define the type in Golang code."""
    return 'type {} {}'.format(typedef.identifier, _express_type(typedef=typedef))


_IMPORTS_TPL = ENV.from_string('''\
{% if imports|length == 0 %}
{% elif imports|length == 1 %}
import "{{ imports[0] }}"{#
#}{% else %}
import (
{% for imprt in imports %}
    "{{ imprt }}"
{% endfor %}
){% endif %}''')


@icontract.ensure(lambda result: not result.endswith('\n'))
@icontract.ensure(lambda import_set, result: len(import_set) != 0 or result == '')
def _state_imports(import_set: Set[str]) -> str:
    """State the imports in Golang code."""
    return _IMPORTS_TPL.render(imports=sorted(import_set))


_TYPES_GO_TPL = ENV.from_string('''\
package {{ package }}

// Automatically generated file by swagger_to. DO NOT EDIT OR APPEND ANYTHING!
{% if imports_code != '' %}

{{ imports_code }}
{% endif %}
{% for typedef in typedefs.values() %}

{% if typedef.description != '' %}
{{ '%s %s'|format(typedef.identifier, typedef.description)|comment }}
{% endif %}
type {{ typedef.identifier }} {{ type_expression[typedef] }}
{% endfor %}

''')


@icontract.ensure(lambda result: result.endswith('\n'), "final newline")
def generate_types_go(package: str, typedefs: Mapping[str, Typedef]) -> str:
    """
    Generate a file which defines all the involved types.

    :param package: name of the package
    :param typedefs: type definitions
    :return: Golang code
    """
    # imports
    import_set = set()  # type: Set[str]
    for typedef in typedefs.values():
        for _, another_typedef in _walk(typedef=typedef, parent=None):
            if isinstance(another_typedef, Primitivedef):
                if another_typedef.type == 'time.Time':
                    import_set.add('time')

    text = _TYPES_GO_TPL.render(
        package=package,
        imports_code=_state_imports(import_set=import_set),
        typedefs=typedefs,
        type_expression={typedef: _express_type(typedef)
                         for typedef in typedefs.values()})

    return swagger_to.indent.reindent(text=text, indention='\t')


_STRING_ARGUMENT_FROM_STRING_TPL = ENV.from_string('''\
{% if is_pointerdef(argument.typedef) %}
val := {{ string_identifier }}
{{ target_identifier }} = &val{#
#}{% else %}
{{ target_identifier }} = {{ string_identifier }}{#
#}{% endif %}''')

_INT_ARGUMENT_FROM_STRING_TPL = ENV.from_string('''\
{
    parsed, err := strconv.ParseInt({{ string_identifier }}, 10, 64)
    if err != nil {
{% set msg = "Parameter '%s': "|format(argument.parameter_name)|escaped_str %}
        http.Error(w, {{ msg }}+err.Error(), http.StatusBadRequest)
        return
    }
    converted := int(parsed)
{% if is_pointerdef(argument.typedef) %}
    {{ target_identifier }} = &converted
{% else %}
    {{ target_identifier }} = converted
{% endif %}
}''')

_INT64_ARGUMENT_FROM_STRING_TPL = ENV.from_string('''\
{
    parsed, err := strconv.ParseInt({{ string_identifier }}, 10, 64)
    if err != nil {
{% set msg = "Parameter '%s': "|format(argument.parameter_name)|escaped_str %}
        http.Error(w, {{ msg }}+err.Error(), http.StatusBadRequest)
        return
    }
    converted := int64(parsed)
{% if is_pointerdef(argument.typedef) %}
    {{ target_identifier }} = &converted
{% else %}
    {{ target_identifier }} = converted
{% endif %}
}''')

_INT32_ARGUMENT_FROM_STRING_TPL = ENV.from_string('''\
{
    parsed, err := strconv.ParseInt({{ string_identifier }}, 10, 32)
    if err != nil {
{% set msg = "Parameter '%s': "|format(argument.parameter_name)|escaped_str %}
        http.Error(w, {{ msg }}+err.Error(), http.StatusBadRequest)
        return
    }
    converted := int32(parsed)
{% if is_pointerdef(argument.typedef) %}
    {{ target_identifier }} = &converted
{% else %}
    {{ target_identifier }} = converted
{% endif %}
}''')

_FLOAT32_ARGUMENT_FROM_STRING_TPL = ENV.from_string('''\
{
    parsed, err := strconv.ParseFloat({{ string_identifier }}, 32)
    if err != nil {
{% set msg = "Parameter '%s': "|format(argument.parameter_name)|escaped_str %}
        http.Error(w, {{ msg }}+err.Error(), http.StatusBadRequest)
        return
    }
    converted := float32(parsed)
{% if is_pointerdef(argument.typedef) %}
    {{ target_identifier }} = &converted
{% else %}
    {{ target_identifier }} = converted
{% endif %}
}''')

_FLOAT64_ARGUMENT_FROM_STRING_TPL = ENV.from_string('''\
{
    parsed, err := strconv.ParseFloat({{ string_identifier }}, 64)
    if err != nil {
{% set msg = "Parameter '%s': "|format(argument.parameter_name)|escaped_str %}
        http.Error(w, {{ msg }}+err.Error(), http.StatusBadRequest)
        return
    }
    converted := float64(parsed)
{% if is_pointerdef(argument.typedef) %}
    {{ target_identifier }} = &converted
{% else %}
    {{ target_identifier }} = converted
{% endif %}
}''')


@icontract.require(lambda string_identifier: string_identifier == string_identifier.strip())
@icontract.ensure(lambda result: not result.endswith('\n'))
def _argument_from_string(argument: Argument, string_identifier: str) -> str:
    """Generate the code to parse an argument from a string."""
    target_identifier = argument.parsing_identifier

    tajp = ''
    if isinstance(argument.typedef, Primitivedef):
        tajp = argument.typedef.type

    elif isinstance(argument.typedef, Pointerdef):
        if isinstance(argument.typedef.pointed, Primitivedef):
            tajp = argument.typedef.pointed.type
    else:
        raise NotImplementedError("Parsing argument from string {!r} of type: {!r}".format(
            string_identifier, type(argument)))

    assert tajp != '', 'Expected tajp to be set in the previous execution path.'

    if tajp == 'string':
        return _STRING_ARGUMENT_FROM_STRING_TPL.render(
            argument=argument, string_identifier=string_identifier, target_identifier=target_identifier)

    elif tajp == 'int':
        return _INT_ARGUMENT_FROM_STRING_TPL.render(
            argument=argument, string_identifier=string_identifier, target_identifier=target_identifier)

    elif tajp == 'int64':
        return _INT64_ARGUMENT_FROM_STRING_TPL.render(
            argument=argument, string_identifier=string_identifier, target_identifier=target_identifier)

    elif tajp == 'int32':
        return _INT32_ARGUMENT_FROM_STRING_TPL.render(
            argument=argument, string_identifier=string_identifier, target_identifier=target_identifier)

    elif tajp == 'float32':
        return _FLOAT32_ARGUMENT_FROM_STRING_TPL.render(
            argument=argument, string_identifier=string_identifier, target_identifier=target_identifier)

    elif tajp == 'float64':
        return _FLOAT64_ARGUMENT_FROM_STRING_TPL.render(
            argument=argument, string_identifier=string_identifier, target_identifier=target_identifier)

    else:
        raise NotImplementedError("Parsing argument from string {!r} of Go type: {!r}".format(string_identifier, tajp))


_ARGUMENT_FROM_BODY_TPL = ENV.from_string('''\
{
    var err error
    r.Body = http.MaxBytesReader(w, r.Body, 1024*1024)
    body, err := ioutil.ReadAll(r.Body)
    if err != nil {
        http.Error(w, "Body unreadable: "+err.Error(), http.StatusBadRequest)
        return
    }

    err = ValidateAgainst{{ argument.json_schema.identifier|capital_camel_case }}Schema(body)
    if err != nil {
        http.Error(w, "Failed to validate against schema: "+err.Error(), http.StatusBadRequest)
        return
    }

    err = json.Unmarshal(body, &{{ argument.parsing_identifier }})
    if err != nil {
{% set msg = "Error JSON-decoding body parameter '%s': "|format(argument.parameter_name)|escaped_str %}
        http.Error(w, {{ msg }}+err.Error(),
            http.StatusBadRequest)
        return
    }
}''')


@icontract.ensure(lambda result: not result.endswith('\n'))
def _argument_from_body(argument: Argument) -> str:
    """Generate the code to parse the argument from a request body."""
    return _ARGUMENT_FROM_BODY_TPL.render(argument=argument)


_WRAPPER_TPL = ENV.from_string('''\
{% set newliner = joiner("XXX") %}
{% set description %}
{{ route.wrapper.identifier }} wraps the path `{{ route.path }}` with the method "{{ route.method }}".
{% if route.description %}

Path description:
{{ route.description }}
{% endif %}
{% endset %}{# /set description #}
{{ description|trim|comment }}
func {{ route.wrapper.identifier }}(h Handler, w http.ResponseWriter, r *http.Request) {
{% if route.handler.arguments %}{# intermediate variables #}
{% if newliner() %}{{ '\n' }}{% endif %}
{% for argument in route.handler.arguments %}
    var {{ argument.parsing_identifier }} {{ express_or_identify_type[argument]|indent|indent }}
{% endfor %}{# /for argument in route.handler.arguments #}
{% endif %}{# /if intermediate variables #}
{% if route.wrapper.header_arguments %}{### Header arguments ###}
    {% if newliner() %}{{ '\n' }}{% endif %}
    hdr := r.Header
    {% for argument in route.wrapper.header_arguments %}

    {% if argument.required %}
    if _, ok := hdr[{{ argument.parameter_name|escaped_str }}]; !ok {
        {% set msg = "Parameter '%s' expected in header"|format(argument.parameter_name)|escaped_str %}
        http.Error(w, {{ msg }}, http.StatusBadRequest)
        return
    }
    {{ argument_from_string(
        argument, "hdr.Get(%s)"|format(argument.parameter_name|escaped_str))|indent }}
    {% else %}
    if _, ok := hdr[{{ argument.parameter_name|escaped_str }}]; ok {
        {{ argument_from_string(
            argument, "hdr.Get(%s)"|format(argument.parameter_name|escaped_str))|indent|indent }}
    }
    {% endif %}{# /if argument.required #}
    {% endfor %}{# /for argument in route.wrapper.header_arguments #}
{% endif %}{# /if header arguments #}
{% if route.wrapper.query_arguments %}{### Query arguments ###}
    {% if newliner() %}{{ '\n' }}{% endif %}
    q := r.URL.Query()
    {% for argument in route.wrapper.query_arguments %}

    {% if argument.required %}
    if _, ok := q[{{ argument.parameter_name|escaped_str }}]; !ok {
        {% set msg = "Parameter '%s' expected in query"|format(argument.parameter_name)|escaped_str %}
        http.Error(w, {{ msg }}, http.StatusBadRequest)
        return
    }
    {{ argument_from_string(
        argument, "q.Get(%s)"|format(argument.parameter_name|escaped_str))|indent }}
    {% else %}
    if _, ok := q[{{ argument.parameter_name|escaped_str }}]; ok {
        {{ argument_from_string(
            argument, "q.Get(%s)"|format(argument.parameter_name|escaped_str))|indent|indent }}
    }
    {% endif %}{# /if argument.required #}
    {% endfor %}{# /for query arguments #}
{% endif %}{# /if query arguments #}
{% if route.wrapper.path_arguments %}{### Path arguments ###}
    {% if newliner() %}{{ '\n' }}{% endif %}
    vars := mux.Vars(r)
    {% for argument in route.wrapper.path_arguments %}

    {% if argument.required %}
    if _, ok := vars[{{ argument.parameter_name|escaped_str }}]; !ok {
    {% set msg = "Parameter '%s' expected in path"|format(argument.parameter_name)|escaped_str %}
        http.Error(w, {{ msg }}, http.StatusBadRequest)
        return
    }
    {{ argument_from_string(argument, "vars[%s]"|format(argument.parameter_name|escaped_str))|indent }}
    {% else %}
    if value, ok := vars[{{ argument.parameter_name|escaped_str }}]; ok {
        {{ argument_from_string(argument, "vars[%s]"|format(argument.parameter_name|escaped_str))|indent|indent }}
    }
    {% endif %}
    {% endfor %}{# /path arguments #}
{% endif %}{# /if path arguments #}
{% if route.wrapper.body_argument is not none %}{### Body argument ###}
    {% if newliner() %}{{ '\n' }}{% endif %}
    {% if route.wrapper.body_argument.required %}
    if r.Body == nil {
    {% set msg = "Parameter '%s' expected in body, but got no body"|format(route.wrapper.body_argument.parameter_name)|escaped_str %}
        http.Error(w, {{ msg }}, http.StatusBadRequest)
        return
    }
    {{ argument_from_body(route.wrapper.body_argument)|indent }}
    {% else %}
    if r.Body != nil {
        {{ argument_from_body(route.wrapper.body_argument)|indent|indent }}
    }
    {% endif %}{# /if route.wrapper.body_argument.required #}
{% endif %}{# /if body argument #}
{% if newliner() %}{{ '\n' }}{% endif %}
{% if not route.handler.arguments %}
    h.{{ route.handler.identifier }}(w, r)
{% else %}
    h.{{ route.handler.identifier }}(w,
        r,
{% for argument in route.handler.arguments %}
        {{ argument.parsing_identifier }}{{ "," if not loop.last else ")" }}
{% endfor %}
{% endif %}
}
''')

_ROUTES_GO_TPL = ENV.from_string('''\
package {{ package }}

// Automatically generated file by swagger_to. DO NOT EDIT OR APPEND ANYTHING!
{% if imports_code != '' %}

{{ imports_code }}
{% endif %}

// SetupRouter sets up a router. If you don't use any middleware, you are good to go.
// Otherwise, you need to maually re-implement this function with your middlewares.
func SetupRouter(h Handler) *mux.Router {
    r := mux.NewRouter()
{% for route in routes %}

    r.HandleFunc(`{{ route.path }}`,
        func(w http.ResponseWriter, r *http.Request) {
            {{ route.wrapper.identifier }}(h, w, r)
        }).Methods({{ route.method|escaped_str }})
{% endfor %}

    return r
}
{% if routes %}
{% for route in routes %}

{{ wrapper_code[route] }}
{% endfor %}

{% endif %}
// Automatically generated file by swagger_to. DO NOT EDIT OR APPEND ANYTHING!

''')


@icontract.ensure(lambda result: result.endswith('\n'), "final new line")
def generate_routes_go(package: str, routes: List[Route]) -> str:
    """
    Generate the file which defines the router and the routes.

    :param package: name of the package
    :param routes: routes that the router will handle.
    :return: Golang code
    """
    # imports
    import_set = {"github.com/gorilla/mux", "net/http"}

    for route in routes:
        for argument in route.handler.arguments:
            if argument.in_what == 'body':
                import_set.add('io/ioutil')
                import_set.add('encoding/json')

            if argument.in_what == 'query' or argument.in_what == 'header':
                tajp = ''
                if isinstance(argument.typedef, Primitivedef):
                    tajp = argument.typedef.type
                elif isinstance(argument.typedef, Pointerdef):
                    if isinstance(argument.typedef.pointed, Primitivedef):
                        tajp = argument.typedef.pointed.type

                if tajp in ['int', 'int32', 'int64', 'float32', 'float64']:
                    import_set.add('strconv')

    imports_code = _state_imports(import_set=import_set)

    wrapper_code = {
        route: _WRAPPER_TPL.render(
            route=route,
            express_or_identify_type={
                argument: _express_or_identify_type(typedef=argument.typedef)
                for route in routes for argument in route.handler.arguments
            },
            argument_from_string=_argument_from_string,
            argument_from_body=_argument_from_body)
        for route in routes
    }

    text = _ROUTES_GO_TPL.render(package=package, imports_code=imports_code, routes=routes, wrapper_code=wrapper_code)

    return swagger_to.indent.reindent(text=text, indention='\t')


_HANDLER_IMPL_GO_TPL = ENV.from_string('''\
package {{ package }}

import (
    "net/http"
    "log"
)

// HandlerImpl implements the Handler.
type HandlerImpl struct {
    LogErr *log.Logger
    LogOut *log.Logger}
{% for route in routes %}

// {{ route.handler.identifier }} implements Handler.{{ route.handler.identifier }}.
{% if not route.handler.arguments %}
func (h *HandlerImpl) {{ route.handler.identifier }}(w http.ResponseWriter,
    r *http.Request) {
{% else %}
func (h *HandlerImpl) {{ route.handler.identifier }}(w http.ResponseWriter,
    r *http.Request,
{% for argument in route.handler.arguments %}
    {{ argument.identifier }} {{ argument_type[argument] }}{{ ',' if not loop.last else ') {' }}
{% endfor %}
{% endif %}{# /if not route.handler.arguments #}
    {% set msg = "Not implemented: %s"|format(route.handler.identifier)|escaped_str %}
    http.Error(w, {{ msg }}, http.StatusInternalServerError)
    h.LogErr.Printf({{ msg }})
}
{% endfor %}{# /routes #}
''')


@icontract.ensure(lambda result: result.endswith('\n'), "final newline")
def generate_handler_impl_go(package: str, routes: List[Route]) -> str:
    """
    Generate a file which implements the handler interface with empty methods.

    :param package: name of the package
    :param routes: that a handler will handle
    :return: Golang code
    """
    text = _HANDLER_IMPL_GO_TPL.render(
        package=package,
        routes=routes,
        argument_type={
            argument: _express_or_identify_type(argument.typedef)
            for route in routes for argument in route.handler.arguments
        })

    return swagger_to.indent.reindent(text=text, indention='\t')


_HANDLER_GO_TPL = ENV.from_string('''\
package {{ package }}

// Automatically generated file by swagger_to. DO NOT EDIT OR APPEND ANYTHING!

import "net/http"

// Handler defines an interface to handling the routes.
type Handler interface {
{% for route in routes %}
    {% if not loop.first %}

    {% endif %}
    {% set handler_description %}
{{ route.handler.identifier }} handles the path `{{ route.path }}` with the method "{{ route.method }}".
{% if route.description %}

Path description:
{{ route.description }}
{% endif %}{# /if route.description #}
    {% endset %}{# /set handler_description #}
    {{ handler_description|trim|comment|indent }}
    {% if not route.handler.arguments %}
    {{ route.handler.identifier }}(w http.ResponseWriter,
        r *http.Request)
    {% else %}
    {{ route.handler.identifier }}(w http.ResponseWriter,
        r *http.Request,
    {% for argument in route.handler.arguments %}
        {{ argument.identifier }} {{ argument_type[argument] }}{{ ',' if not loop.last else ')' }}
    {% endfor %}
    {% endif %}{# /if not route.handler.arguments #}
{% endfor %}
}

// Automatically generated file by swagger_to. DO NOT EDIT OR APPEND ANYTHING!

''')


@icontract.ensure(lambda result: result.endswith('\n'), "final newline")
def generate_handler_go(package: str, routes: List[Route]) -> str:
    """
    Generate a file which defines the handler interface.

    :param package: name of the package
    :param routes: that a handler will handle
    :return: Golang code
    """
    text = _HANDLER_GO_TPL.render(
        package=package,
        routes=routes,
        argument_type={
            argument: _express_or_identify_type(argument.typedef)
            for route in routes for argument in route.handler.arguments
        })

    return swagger_to.indent.reindent(text=text, indention='\t')


_JSON_SCHEMAS_GO_TPL = ENV.from_string('''\
{# This template must be indented with tabs since we need to include the schema as text and hence can not re-indent
   since re-indention . #}
package {{ package }}

// Automatically generated file by swagger_to. DO NOT EDIT OR APPEND ANYTHING!

{% if not schemas %}
// No schemas are defined in the swagger.
{% else %}
import (
	"errors"
	"fmt"
	"github.com/xeipuuv/gojsonschema"
)

func mustNewJSONSchema(text string, name string) *gojsonschema.Schema {
	loader := gojsonschema.NewStringLoader(text)
	schema, err := gojsonschema.NewSchema(loader)
	if err != nil {
		panic(fmt.Sprintf("failed to load JSON Schema %#v: %s", text, err.Error()))
	}
	return schema
}
{% for schema in schemas.values() %}

var jsonSchema{{ schema.identifier|capital_camel_case }}Text = `{{ schema.text|replace('`', '` + "`" + `') }}`
{% endfor %}
{% for schema in schemas.values() %}

var jsonSchema{{ schema.identifier|capital_camel_case }} = mustNewJSONSchema(
	jsonSchema{{ schema.identifier|capital_camel_case }}Text,
	{{ schema.identifier|escaped_str }})
{% endfor %}
{% for schema in schemas.values() %}

{% set validateFuncName = "ValidateAgainst%sSchema"|format(schema.identifier|capital_camel_case) %}
// {{ validateFuncName }} validates a message coming from the client against {{ schema.identifier }} schema.
func {{ validateFuncName }}(bb []byte) error {
	loader := gojsonschema.NewStringLoader(string(bb))
	result, err := jsonSchema{{ schema.identifier|capital_camel_case }}.Validate(loader)
	if err != nil {
		return err
	}

	if result.Valid() {
		return nil
	}

	msg := ""
	for i, valErr := range result.Errors() {
		if i > 0 {
			msg += ", "
		}
		msg += valErr.String()
	}
	return errors.New(msg)
}
{% endfor %}
{% endif %}{# /if not schemas #}

// Automatically generated file by swagger_to. DO NOT EDIT OR APPEND ANYTHING!

''')


@icontract.ensure(lambda result: result.endswith('\n'), "final newline")
def generate_json_schemas_go(package: str, routes: List[Route], typedefs: MutableMapping[str, Typedef]) -> str:
    """
    Represent the definitions as json schemas and hard-codes them as strings in Go.

    It is assumed that the Swagger definitions already represent a subset of JSON Schema.
    This is theoretically not the case (some formats are swagger-only), but in most cases
    the literal translation should work.

    :param package: package name
    :param routes: needed to generate the parameter schemas if they are not already defined in the definitions
    :param typedefs: type definitions to generate the schemas for
    :return: Golang code
    """
    schemas = collections.OrderedDict()  # type: MutableMapping[str, JsonSchema]

    for route in routes:
        if route.wrapper.body_argument is not None:
            if not route.wrapper.body_argument.json_schema.identifier in schemas:
                schemas[route.wrapper.body_argument.json_schema.identifier] = route.wrapper.body_argument.json_schema

    for typedef_id, typedef in typedefs.items():
        if typedef.json_schema is None:
            raise AssertionError("Missing JSON schema for typedef: {!r}".format(typedef_id))

        # Assume the typedef identifiers are unique accross routes and typedefs.
        if typedef.json_schema.identifier not in schemas:
            schemas[typedef.json_schema.identifier] = typedef.json_schema

    return _JSON_SCHEMAS_GO_TPL.render(package=package, schemas=schemas)



#!/usr/bin/env python3
"""Generate code for an TypeScript client with Angular5."""

from typing import MutableMapping, Optional, TextIO, List  # pylint: disable=unused-import

import collections
import icontract

import swagger_to
import swagger_to.intermediate
import swagger_to.swagger

INDENT = ' ' * 4


class Typedef:
    """Represent the type definition in Typescript."""

    def __init__(self) -> None:
        """Initialize with defaults."""
        self.description = ''
        self.identifier = ''

    def __str__(self) -> str:
        """Represent the type definition as its class name and the identifier."""
        return '{}({})'.format(self.__class__.__name__, self.identifier)


class Booleandef(Typedef):
    """Represent a Typescript boolean."""

    pass


class Numberdef(Typedef):
    """Represent a Typescript number."""

    pass


class Stringdef(Typedef):
    """Represent a Typescript string."""

    pass


class Arraydef(Typedef):
    """Represents a Typescript array."""

    def __init__(self) -> None:
        """Initialize with defaults."""
        self.items = None  # type: Optional[Typedef]
        super().__init__()

    def __str__(self) -> str:
        """Represent the array as its class name, identifier and the recursive representation of the items type."""
        return '{}({}, items: {})'.format(self.__class__.__name__, self.identifier, self.items)


class Mapdef(Typedef):
    """Represents a Typescript map."""

    def __init__(self) -> None:
        """Initialize with defaults."""
        self.values = None  # type: Optional[Typedef]
        super().__init__()

    def __str__(self) -> str:
        """Represent the array as its class name, identifier and the recursive representation of the values type."""
        return '{}({}, values: {})'.format(self.__class__.__name__, self.identifier, self.values)


class Property:
    """Represents a Typescript class property."""

    def __init__(self) -> None:
        """Initialize with defaults."""
        self.name = ''
        self.description = ''
        self.typedef = None  # type: Optional[Typedef]
        self.required = False

    def __str__(self) -> str:
        """Represent the property as its name and the type."""
        return '{}: {}'.format(self.name, self.typedef)


class Classdef(Typedef):
    """Represent a Typescript class."""

    def __init__(self) -> None:
        """Initialize with defaults."""
        self.properties = collections.OrderedDict()  # type: MutableMapping[str, Property]
        super().__init__()

    def __str__(self) -> str:
        """Represent the class as its class name, identifier and the list of properties."""
        return '{}({}, properties: {})'.format(self.__class__.__name__, self.identifier,
                                               ', '.join([str(val) for val in self.properties.values()]))


class Parameter:
    """Represent a parameter of a request function in Typescript."""

    def __init__(self) -> None:
        """Initialize with defaults."""
        self.name = ''
        self.typedef = None  # type: Optional[Typedef]
        self.required = False


class Response:
    """Represent a response from a request in Typescript."""

    def __init__(self) -> None:
        """Initialize with defaults."""
        self.code = ''
        self.typedef = None  # type: Optional[Typedef]
        self.description = ''


class Request:
    """Represent a request function of the client."""

    # pylint: disable=too-many-instance-attributes
    def __init__(self) -> None:
        """Initialize with defaults."""
        self.operation_id = ''
        self.path = ''
        self.method = ''
        self.description = ''

        # the union of all parameters
        self.parameters = []  # type: List[Parameter]

        # parameters split into categories
        self.body_parameter = None  # type: Optional[Parameter]
        self.query_parameters = []  # type: List[Parameter]
        self.path_parameters = []  # type: List[Parameter]

        self.responses = collections.OrderedDict()  # type: MutableMapping[str, Response]


def _to_typedef(intermediate_typedef: swagger_to.intermediate.Typedef) -> Typedef:
    """
    Translate the type definition in intermediate representation to Typescript.

    :param intermediate_typedef: type definition in intermediate representation
    :return: type definition in Typescript representation
    """
    typedef = None  # type: Optional[Typedef]

    if isinstance(intermediate_typedef, swagger_to.intermediate.Primitivedef):
        if intermediate_typedef.type == 'boolean':
            typedef = Booleandef()
        elif intermediate_typedef.type == 'number' or intermediate_typedef.type == 'integer':
            typedef = Numberdef()
        elif intermediate_typedef.type == 'string':
            typedef = Stringdef()
        else:
            raise NotImplementedError("Converting intermediate type to Typescript is not supported: {}".format(
                intermediate_typedef.type))

    elif isinstance(intermediate_typedef, swagger_to.intermediate.Arraydef):
        typedef = Arraydef()
        typedef.items = _to_typedef(intermediate_typedef=intermediate_typedef.items)

    elif isinstance(intermediate_typedef, swagger_to.intermediate.Mapdef):
        typedef = Mapdef()
        typedef.values = _to_typedef(intermediate_typedef=intermediate_typedef.values)

    elif isinstance(intermediate_typedef, swagger_to.intermediate.Objectdef):
        typedef = Classdef()

        for intermediate_prop in intermediate_typedef.properties.values():
            prop = Property()
            prop.description = intermediate_prop.description
            prop.name = intermediate_prop.name
            prop.typedef = _to_typedef(intermediate_typedef=intermediate_prop.typedef)
            prop.required = intermediate_prop.required

            typedef.properties[prop.name] = prop
    else:
        raise NotImplementedError("Converting intermediate typedef to Typescript is not supported: {!r}".format(
            type(intermediate_typedef)))

    typedef.description = intermediate_typedef.description
    typedef.identifier = intermediate_typedef.identifier

    assert typedef is not None

    return typedef


def to_typedefs(
        intermediate_typedefs: MutableMapping[str, swagger_to.intermediate.Typedef]) -> MutableMapping[str, Typedef]:
    """
    Translate type definitions from intermediate representation to Typescript.

    :param intermediate_typedefs: table of type definitions in intermediate representation
    :return: table of type definitions in Typescript representation
    """
    typedefs = collections.OrderedDict()  # type: MutableMapping[str, Typedef]

    for intermediate_typedef in intermediate_typedefs.values():
        assert intermediate_typedef is not None

        typedef = _to_typedef(intermediate_typedef=intermediate_typedef)
        typedefs[typedef.identifier] = typedef

    return typedefs


def _anonymous_or_get_typedef(intermediate_typedef: swagger_to.intermediate.Typedef,
                              typedefs: MutableMapping[str, Typedef]) -> Typedef:
    """
    Get the Typescript type definition from the table of Typescript type definitions by its identifier.

    If the type definition has no identifier, it is translated on the spot
    to the corresponding Typescript type definition.

    :param intermediate_typedef: intermediate representation of the type definition
    :param typedefs: table of type definitions in Typescript representation
    :return: type definition in Typescript representation
    """
    if intermediate_typedef.identifier:
        if intermediate_typedef.identifier not in typedefs:
            raise KeyError("Intermediate typedef not found in the translated typedefs: {!r}".format(
                intermediate_typedef.identifier))

        return typedefs[intermediate_typedef.identifier]

    return _to_typedef(intermediate_typedef=intermediate_typedef)


def _to_parameter(intermediate_parameter: swagger_to.intermediate.Parameter,
                  typedefs: MutableMapping[str, Typedef]) -> Parameter:
    """
    Translate an endpoint parameter from the intermediate to a Typescript representation.

    :param intermediate_parameter: intermediate representation of a parameter
    :param typedefs: table of type definitions in Typescript representation
    :return: Typescript representation of the parameter
    """
    param = Parameter()
    param.name = intermediate_parameter.name
    param.typedef = _anonymous_or_get_typedef(intermediate_typedef=intermediate_parameter.typedef, typedefs=typedefs)
    param.required = intermediate_parameter.required
    return param


def _to_response(intermediate_response: swagger_to.intermediate.Response,
                 typedefs: MutableMapping[str, Typedef]) -> Response:
    """
    Translate an endpoint response from the intermediate to a Typescript representation.

    :param intermediate_response: intermediate representation of a response
    :param typedefs: table of type definitions in Typescript representation
    :return: Typescript representation of the response
    """
    resp = Response()
    resp.code = intermediate_response.code
    resp.typedef = None if intermediate_response.typedef is None else \
        _anonymous_or_get_typedef(intermediate_typedef=intermediate_response.typedef, typedefs=typedefs)
    resp.description = intermediate_response.description
    return resp


def _to_request(endpoint: swagger_to.intermediate.Endpoint, typedefs: MutableMapping[str, Typedef]) -> Request:
    """
    Translate an endpoint from an intermediate representation to a Typescript client request function.

    :param endpoint: intermediate representation of the endpoint
    :param typedefs: table of type definitions in Typescript representation
    :return: Typescript representation of the client request function
    """
    if endpoint.produces != ['application/json']:
        raise ValueError("Can not translate an endpoint to Typescript client "
                         "which does not produces strictly application/json: {} {}".format(
                             endpoint.path, endpoint.method))

    req = Request()
    req.description = endpoint.description
    req.method = endpoint.method
    req.operation_id = endpoint.operation_id
    req.path = endpoint.path

    for intermediate_param in endpoint.parameters:
        param = _to_parameter(intermediate_parameter=intermediate_param, typedefs=typedefs)

        if intermediate_param.in_what == 'body':
            if req.body_parameter is not None:
                raise KeyError("Duplicate body parameters in an endpoint: {} {}".format(
                    req.body_parameter.name, intermediate_param.name))

            req.body_parameter = param
        elif intermediate_param.in_what == 'query':
            req.query_parameters.append(param)
        elif intermediate_param.in_what == 'path':
            req.path_parameters.append(param)
        else:
            raise NotImplementedError("Unsupported parameter 'in' to Typescript translation: {}".format(
                intermediate_param.in_what))

        req.parameters.append(param)

    # parameters are sorted so that first come the required ones; Typescript requires the optional ones to come
    # at the end.
    req.parameters.sort(key=lambda param: not param.required)

    for code, intermediate_resp in endpoint.responses.items():
        req.responses[code] = _to_response(intermediate_response=intermediate_resp, typedefs=typedefs)

    return req


@icontract.ensure(
    lambda endpoints, result:
    len(result) == len([endpoint for endpoint in endpoints if endpoint.produces == ['application/json']]))
def to_requests(endpoints: List[swagger_to.intermediate.Endpoint],
                typedefs: MutableMapping[str, Typedef]) -> List[Request]:
    """
    Translate endpoints from intermediate representation to Typescript client request functions.

    Endpoints which do not produce strictly 'application/json' are ignored.

    :param endpoints: intermediate representation of the endpoints
    :param typedefs: table of type definitions in Typescript representation
    :return: Typescript representation of client's request functions corresponding to the endpoints
    """
    requests = []  # type: List[Request]
    for endpoint in endpoints:
        if endpoint.produces != ['application/json']:
            continue

        requests.append(_to_request(endpoint=endpoint, typedefs=typedefs))

    return requests


def _write_header(fid: TextIO) -> None:
    """
    Write the header of the client file.

    :param fid: target
    :return:
    """
    fid.write("// Automatically generated file by swagger_to. DO NOT EDIT OR APPEND ANYTHING!\n\n")

    fid.write("import { Injectable } from '@angular/core';\n"
              "import { Http } from '@angular/http';\n"
              "import { HttpErrorResponse } from '@angular/common/http';\n"
              "import { Observable } from 'rxjs/Rx';\n"
              "import { RequestOptions } from '@angular/http';\n\n")


def _write_footer(fid: TextIO) -> None:
    """
    Write the footer of the client file.

    :param fid: target
    :return:
    """
    fid.write("\n\n// Automatically generated file by swagger_to. DO NOT EDIT OR APPEND ANYTHING!\n")


def _type_expression(typedef: Typedef, path: Optional[str] = None) -> str:
    """
    Translate the type definition in Typescript representation to a type expression as Typescript code.

    :param typedef: Typescript representation of the type definition
    :param path: path in the Swagger spec
    :return: Typescript code
    """
    if isinstance(typedef, Booleandef):
        return 'boolean'
    elif isinstance(typedef, Numberdef):
        return 'number'
    elif isinstance(typedef, Stringdef):
        return 'string'
    elif isinstance(typedef, Arraydef):
        if typedef.items is None:
            raise ValueError('Unexpected None items in typedef: {!r}'.format(typedef.identifier))

        return 'Array<' + _type_expression(typedef=typedef.items, path=str(path) + '.items') + ">"
    elif isinstance(typedef, Mapdef):
        if typedef.values is None:
            raise ValueError('Unexpected None values in typedef: {!r}'.format(typedef.identifier))

        return 'Map<string, ' + _type_expression(typedef=typedef.values, path=str(path) + '.values') + ">"
    elif isinstance(typedef, Classdef):
        if typedef.identifier == '':
            raise NotImplementedError(
                "Translating an anonymous class to a Typescript type expression is not supported: {}".format(path))

        return typedef.identifier
    else:
        raise NotImplementedError("Translating the typedef to a type expression is not supported: {!r}: {}".format(
            type(typedef), path))


def _write_description(description: str, indent: str, fid: TextIO) -> None:
    """
    Write a description as // comment block.

    :param description: to be written
    :param indent: indention as prefix to each line
    :param fid: target
    :return:
    """
    lines = description.strip().splitlines()
    for i, line in enumerate(lines):
        rstripped = line.rstrip()
        if len(rstripped) > 0:
            fid.write(indent + '// {}'.format(rstripped))
        else:
            fid.write(indent + '//')

        if i < len(lines) - 1:
            fid.write('\n')


def _write_type_definition(typedef: Typedef, fid: TextIO) -> None:
    """
    Write the type definition in the Typescript code.

    :param typedef: to be declared
    :param fid: target
    :return:
    """
    if typedef.identifier == '':
        raise ValueError("Expected a typedef with an identifier, but got a typedef with an empty identifier.")

    if typedef.description:
        _write_description(description=typedef.description, indent='', fid=fid)
        fid.write('\n')

    if isinstance(typedef, Classdef):
        fid.write("export interface {} {{\n".format(typedef.identifier))
        for i, prop in enumerate(typedef.properties.values()):
            if i > 0:
                fid.write('\n')

            if prop.description:
                _write_description(description=prop.description, indent=INDENT, fid=fid)
                fid.write("\n")

            if prop.typedef is None:
                raise ValueError('Unexpected typedef of prop {!r} in class {!r}'.format(prop.name, typedef.identifier))

            type_expr = _type_expression(typedef=prop.typedef, path='{}.{}'.format(typedef.identifier, prop.name))

            if not prop.required:
                fid.write(INDENT + '{}?: {};\n'.format(prop.name, type_expr))
            else:
                fid.write(INDENT + '{}: {};\n'.format(prop.name, type_expr))

        fid.write("}")
    else:
        fid.write("type {} = {};".format(typedef.identifier, _type_expression(typedef=typedef,
                                                                              path=typedef.identifier)))


def _write_type_definitions(typedefs: MutableMapping[str, Typedef], fid: TextIO) -> None:
    """
    Write all type definitions as Typescript code.

    :param typedefs: type definitions to be declared
    :param fid: target
    :return:
    """
    for i, typedef in enumerate(typedefs.values()):
        if i > 0:
            fid.write('\n\n')

        _write_type_definition(typedef=typedef, fid=fid)


def _to_string_expression(typedef: Typedef, variable: str) -> str:
    """
    Wrap the variable with .toString() if necessary.

    :param typedef: type definition of the variable in Typescript representation
    :param variable: variable to be converted to string in the generated code
    :return: Typescript code
    """
    if isinstance(typedef, Stringdef):
        return variable

    return '{}.toString()'.format(variable)


def _write_request(request: Request, fid: TextIO) -> None:
    """
    Generate the code of the request function.

    :param request: function definition
    :param fid: target
    :return:
    """
    # pylint: disable=too-many-locals
    # pylint: disable=too-many-statements
    # pylint: disable=too-many-branches
    description = 'Sends a request to the endpoint: {} {}'.format(request.path, request.method)
    if request.description:
        description += '\n\n' + request.description
    _write_description(description, INDENT, fid)
    fid.write('\n')

    prefix = INDENT + 'public {}('.format(request.operation_id)

    args = []  # type: List[str]
    for param in request.parameters:
        if param.typedef is None:
            raise ValueError('Unexpected None typedef in param {!r} of request {!r}'.format(
                param.name, request.operation_id))

        args.append('{}{}: {}'.format(
            param.name, '?' if not param.required else '', _type_expression(typedef=param.typedef)))

    return_type = ''
    if '200' in request.responses:
        resp = request.responses['200']
        if resp.typedef is not None:
            return_type = _type_expression(typedef=resp.typedef)

    if not return_type:
        return_type = 'any'

    suffix = '): Observable<{} | HttpErrorResponse> {{'.format(return_type)

    line = prefix + ', '.join(args) + suffix
    if len(line) <= 120:
        fid.write(line)
    else:
        fid.write(prefix)
        fid.write('\n')

        for i, arg in enumerate(args):
            fid.write(3 * INDENT + arg)

            if i < len(args) - 1:
                fid.write(',\n')

        fid.write(suffix)
    fid.write('\n')

    name_to_parameters = {param.name: param for param in request.parameters}

    rel_path = request.path[1:] if request.path.startswith('/') else request.path

    # path parameters
    token_pth = swagger_to.tokenize_path(path=rel_path)

    if not token_pth.parameter_to_token_indices and not request.query_parameters:
        fid.write(INDENT * 2 + 'const url = this.url_prefix + "{}";\n'.format(rel_path))
    else:
        if not token_pth.parameter_to_token_indices:
            fid.write(INDENT * 2 + 'let url = this.url_prefix + "{}";'.format(rel_path))
        else:
            fid.write(INDENT * 2 + 'let url = this.url_prefix;')
            for i, tkn in enumerate(token_pth.tokens):
                fid.write("\n")

                if i in token_pth.token_index_to_parameter:
                    param_name = token_pth.token_index_to_parameter[i]
                    param = name_to_parameters[param_name]

                    if param.typedef is None:
                        raise ValueError('Unexpected None typedef in param {!r} and request {!r}'.format(
                            param.name, request.operation_id))

                    fid.write(INDENT * 2 + 'url += encodeURIComponent({});'.format(
                        _to_string_expression(typedef=param.typedef, variable=param.name)))
                else:
                    fid.write(INDENT * 2 + 'url += encodeURIComponent("{}");'.format(
                        tkn.replace('\\', '\\\\').replace('\n', '\\n').replace('\r', '\\r').replace('\t', '\\t')))

    if request.path_parameters and request.query_parameters:
        fid.write("\n")

    # query parameters
    if request.query_parameters:
        fid.write('\n')
        fid.write(INDENT * 2 + 'url += "?";\n')

        for i, param in enumerate(request.query_parameters):
            amp = ''  # used to concatenate query parameters
            if i > 0:
                fid.write("\n\n")
                amp = '&'

            if param.typedef is None:
                raise ValueError('Unexpected None typedef in param {!r} and request {!r}'.format(
                    param.name, request.operation_id))

            if param.required:
                fid.write(INDENT * 2 + 'url += "{}{}=" + encodeURIComponent({});'.format(
                    amp, param.name, _to_string_expression(typedef=param.typedef, variable=param.name)))
            else:
                fid.write(INDENT * 2 + 'if ({}) {{\n'.format(param.name))
                fid.write(INDENT * 3 + 'url += "{}{}=" + encodeURIComponent({});\n'.format(
                    amp, param.name, _to_string_expression(typedef=param.typedef, variable=param.name)))
                fid.write(INDENT * 2 + '}')

        fid.write('\n')

    fid.write('\n')

    mth = request.method.lower()
    if request.body_parameter is not None:
        if request.body_parameter.required:
            fid.write(INDENT * 2 + 'let observable = this.http.request(url, \n')
            fid.write(INDENT * 3 + 'new RequestOptions({{method: "{0}", body: JSON.stringify({1})}}));\n'.format(
                mth, request.body_parameter.name))
        else:
            fid.write(INDENT * 2 + 'let observable: Observable<any>;\n')
            fid.write(INDENT * 2 + 'if ({}) {{\n'.format(request.body_parameter.name))
            fid.write(INDENT * 3 + 'this.http.request(url, \n')
            fid.write(INDENT * 4 + 'new RequestOptions({{method: "{0}", body: JSON.stringify({1})}}));\n'.format(
                mth, request.body_parameter.name))
            fid.write(INDENT * 2 + '} else {\n')
            fid.write(INDENT * 3 + 'observable = this.http.request(url, '
                      'new RequestOptions({{method: "{0}"}}));\n'.format(mth))
            fid.write(INDENT * 2 + '}\n')
    else:
        fid.write(INDENT * 2 + 'let observable = this.http.{}(url);\n'.format(mth))

    return_var = 'observable'
    if return_type != 'any':
        fid.write(
            INDENT * 2 + 'let typed_observable = observable.map(res => (res.json() as {}));\n'.format(return_type))
        return_var = 'typed_observable'

    fid.write(INDENT * 2 + 'if (this.on_error) {\n')
    fid.write(INDENT * 3 + 'return {0}.catch(err => this.on_error(err))\n'.format(return_var))
    fid.write(INDENT * 2 + '}\n')
    fid.write(INDENT * 2 + 'return {};\n'.format(return_var))

    fid.write(INDENT + '}')


def _write_client(requests: List[Request], fid: TextIO) -> None:
    """
    Generate the client.

    :param requests: translated request functions
    :param fid: target
    :return:
    """
    fid.write("@Injectable()\n")
    fid.write("export class RemoteCaller {\n")
    fid.write(INDENT + "public url_prefix: string;\n")
    fid.write(INDENT + "public on_error: (error: HttpErrorResponse) => Observable<HttpErrorResponse> | null;\n")
    fid.write("\n")
    fid.write(INDENT + "constructor(private http: Http) {\n")
    fid.write(INDENT * 2 + 'this.url_prefix = "";\n')
    fid.write(INDENT * 2 + 'this.on_error = null;\n')
    fid.write(INDENT + '}\n\n')
    fid.write(INDENT + "public set_url_prefix(url_prefix: string) {\n")
    fid.write(INDENT * 2 + "this.url_prefix = url_prefix;\n")
    fid.write(INDENT + '}\n\n')
    fid.write(INDENT + "public set_on_error("
              "on_error: (error: HttpErrorResponse) => Observable<HttpErrorResponse> | null) {\n")
    fid.write(INDENT * 2 + "this.on_error = on_error;\n")
    fid.write(INDENT + '}')

    for request in requests:
        fid.write('\n\n')
        _write_request(request=request, fid=fid)

    fid.write("\n}")


def write_client_ts(typedefs: MutableMapping[str, Typedef], requests: List[Request], fid: TextIO) -> None:
    """
    Generate the file with the client code.

    :param typedefs: table of type definitions in Typescript representation
    :param requests: request functions in Typescript representation
    :param fid: target
    :return:
    """
    _write_header(fid=fid)

    if typedefs:
        _write_type_definitions(typedefs=typedefs, fid=fid)

    if requests and typedefs:
        fid.write('\n\n')

    if requests:
        _write_client(requests=requests, fid=fid)
        _write_footer(fid=fid)



#!/usr/bin/env python3
"""Read a correct swagger file and produce Go server code."""
import argparse
import pathlib
import sys

import swagger_to.go_server
import swagger_to.intermediate
import swagger_to.swagger


def main() -> None:
    """Execute the main routine."""
    # pylint: disable=too-many-locals
    parser = argparse.ArgumentParser("Reads a correct swagger file and produces Go code")
    parser.add_argument("--swagger_path", help="path to the swagger file", required=True)
    parser.add_argument("--outdir", help="path to the output directory", required=True)
    parser.add_argument("--no_samples", help="if set, do not generate sample files", action="store_true")
    parser.add_argument("--force", help="overwrite existing files", action="store_true")
    args = parser.parse_args()

    assert isinstance(args.force, bool)
    assert isinstance(args.no_samples, bool)
    assert isinstance(args.outdir, str)
    assert isinstance(args.swagger_path, str)

    force = bool(args.force)
    no_samples = bool(args.no_samples)
    swagger_path = pathlib.Path(args.swagger_path)
    outdir = pathlib.Path(args.outdir)

    if not swagger_path.exists():
        print("Swagger file does not exist: {}".format(swagger_path), file=sys.stderr)
        sys.exit(1)

    outdir.mkdir(parents=True, exist_ok=True)

    if not force:
        for fname in ['main.go', 'handlers.go', 'types.go', 'routes.go', 'jsonschemas.go']:
            pth = outdir / fname
            if pth.exists():
                print("File exists, but --force was not specified: {!r}".format(pth), file=sys.stderr)
                sys.exit(1)

    swagger, errs = swagger_to.swagger.parse_yaml_file(path=swagger_path.as_posix())
    if errs:
        raise AssertionError("Errors in {}:\n{}".format(swagger_path, "\n".join(errs)))

    intermediate_typedefs = swagger_to.intermediate.to_typedefs(swagger=swagger)
    intermediate_params = swagger_to.intermediate.to_parameters(swagger=swagger, typedefs=intermediate_typedefs)

    endpoints = swagger_to.intermediate.to_endpoints(
        swagger=swagger, typedefs=intermediate_typedefs, params=intermediate_params)

    go_typedefs = swagger_to.go_server.to_typedefs(intermediate_typedefs=intermediate_typedefs)
    go_routes = swagger_to.go_server.to_routes(endpoints=endpoints, typedefs=go_typedefs)

    package = swagger.name

    (outdir / 'types.go').write_text(swagger_to.go_server.generate_types_go(package=package, typedefs=go_typedefs))

    (outdir / 'routes.go').write_text(swagger_to.go_server.generate_routes_go(package=package, routes=go_routes))

    (outdir / 'handler.go').write_text(swagger_to.go_server.generate_handler_go(package=package, routes=go_routes))

    if not no_samples:
        (outdir / 'handler_impl.go.sample').write_text(
            swagger_to.go_server.generate_handler_impl_go(package=package, routes=go_routes))

    (outdir / 'jsonschemas.go').write_text(
        swagger_to.go_server.generate_json_schemas_go(package=package, routes=go_routes, typedefs=go_typedefs))

    print("Generated go server code in: {}".format(outdir))


if __name__ == "__main__":
    main()



"""Re-indent the code."""
import re
import textwrap
from typing import List  # pylint: disable=unused-import

_SPACE4_RE = re.compile('^([ ]{4})+')


def reindent(text: str, level: int = 0, indention: str = ' ' * 4) -> str:
    r"""
    Strip the prefix indentation, parse indention as 4 spaces and re-indent according to ``level``.

    >>> result = reindent(text='''\
    ...     test me:
    ...         again
    ...             and again
    ...     ''', indention='|')
    >>> assert result == ('test me:\n'
    ...                   '|again\n'
    ...                   '||and again\n')

    >>> result = reindent(text='''\
    ...     test me:
    ...         again
    ...             and again
    ...     ''', level=1, indention='|')
    >>> assert result == ('|test me:\n'
    ...                   '||again\n'
    ...                   '|||and again\n')

    :param text: to be re-indented
    :param level: indention level
    :return: re-indented text
    """
    text = textwrap.dedent(text)

    lines = text.splitlines(keepends=True)
    result_lines = []  # type: List[str]
    for line in lines:
        mtch = _SPACE4_RE.match(line)
        if mtch:
            _, end = mtch.span()
            spaces = end
            assert spaces % 4 == 0, "Expected to match indention at 4 spaces, but got spaces == {}".format(spaces)

            result_lines.append(indention * int(spaces / 4 + level) + line[end:])

        else:
            result_lines.append(indention * level + line)

    return ''.join(result_lines)
